import os
import json
from geopandas import GeoDataFrame
import pytest
from network_wrangler import RoadwayNetwork
from network_wrangler import ProjectCard
import time
import numpy as np
import pandas as pd
from network_wrangler import offset_lat_lon
from network_wrangler import haversine_distance
import networkx as nx

pd.set_option("display.max_rows", 500)
pd.set_option("display.max_columns", 500)
pd.set_option("display.width", 50000)

"""
Run just the tests labeled basic using `pytest -m roadway`
To run with print statments, use `pytest -s -m roadway`
"""

STPAUL_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "examples", "stpaul"
)
STPAUL_SHAPE_FILE = os.path.join(STPAUL_DIR, "shape.geojson")
STPAUL_LINK_FILE = os.path.join(STPAUL_DIR, "link.json")
STPAUL_NODE_FILE = os.path.join(STPAUL_DIR, "node.geojson")

SMALL_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "examples", "single"
)
SMALL_SHAPE_FILE = os.path.join(SMALL_DIR, "shape.geojson")
SMALL_LINK_FILE = os.path.join(SMALL_DIR, "link.json")
SMALL_NODE_FILE = os.path.join(SMALL_DIR, "node.geojson")


def _read_stpaul_net():
    net = RoadwayNetwork.read(
        link_file=STPAUL_LINK_FILE,
        node_file=STPAUL_NODE_FILE,
        shape_file=STPAUL_SHAPE_FILE,
        fast=True,
    )
    return net


SCRATCH_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "scratch"
)


def _read_stpaul_net():
    net = RoadwayNetwork.read(
        link_file=STPAUL_LINK_FILE,
        node_file=STPAUL_NODE_FILE,
        shape_file=STPAUL_SHAPE_FILE,
        fast=True,
    )
    return net


@pytest.mark.roadway
@pytest.mark.travis
def test_roadway_read_write(request):
    print("\n--Starting:", request.node.name)

    out_prefix = "t_readwrite"
    out_shape_file = os.path.join(SCRATCH_DIR, out_prefix + "_" + "shape.geojson")
    out_link_file = os.path.join(SCRATCH_DIR, out_prefix + "_" + "link.json")
    out_node_file = os.path.join(SCRATCH_DIR, out_prefix + "_" + "node.geojson")

    time0 = time.time()

    net = RoadwayNetwork.read(
        link_file=SMALL_LINK_FILE,
        node_file=SMALL_NODE_FILE,
        shape_file=SMALL_SHAPE_FILE,
        fast=True,
    )
    time1 = time.time()
    print("Writing to: {}".format(SCRATCH_DIR))
    net.write(filename=out_prefix, path=SCRATCH_DIR)
    time2 = time.time()
    net_2 = RoadwayNetwork.read(
        link_file=out_link_file, node_file=out_node_file, shape_file=out_shape_file
    )
    time3 = time.time()

    read_time1 = time1 - time0
    read_time2 = time3 - time2
    write_time = time2 - time1

    print("TIME, read (w/out valdiation, with): {},{}".format(read_time1, read_time2))
    print("TIME, write:{}".format(write_time))
    """
    # right now don't have a good way of ignoring differences in rounding
    with open(shape_file, 'r') as s1:
        og_shape = json.loads(s1.read())
        #og_shape.replace('\r', '').replace('\n', '').replace(' ','')
    with open(os.path.join('scratch','t_readwrite_shape.geojson'), 'r')  as s2:
        new_shape = json.loads(s2.read())
        #new_shape.replace('\r', '').replace('\n', '').replace(' ','')
    assert(og_shape==new_shape)
    """


@pytest.mark.roadway
@pytest.mark.travis
def test_quick_roadway_read_write(request):
    print("\n--Starting:", request.node.name)

    out_prefix = "t_readwrite"
    out_shape_file = os.path.join(SCRATCH_DIR, out_prefix + "_" + "shape.geojson")
    out_link_file = os.path.join(SCRATCH_DIR, out_prefix + "_" + "link.json")
    out_node_file = os.path.join(SCRATCH_DIR, out_prefix + "_" + "node.geojson")
    net = RoadwayNetwork.read(
        link_file=SMALL_LINK_FILE,
        node_file=SMALL_NODE_FILE,
        shape_file=SMALL_SHAPE_FILE,
        fast=True,
    )
    net.write(filename=out_prefix, path=SCRATCH_DIR)
    net_2 = RoadwayNetwork.read(
        link_file=out_link_file, node_file=out_node_file, shape_file=out_shape_file
    )
    print("--Finished:", request.node.name)


@pytest.mark.parametrize(
    "selection",
    [
        {
            "link": [{"name": ["6th", "Sixth", "sixth"]}],
            "A": {"osm_node_id": "187899923"},
            "B": {"osm_node_id": "187865924"},
            "answer": ["187899923", "187858777", "187923585", "187865924"],
        },
        {
            "link": [{"name": ["6th", "Sixth", "sixth"]}],
            "A": {"osm_node_id": "187899923"},  # start searching for segments at A
            "B": {"osm_node_id": "187942339"},
        },
        {
            "link": [{"name": ["6th", "Sixth", "sixth"]}, {"lanes": [1, 2]}],
            "A": {"osm_node_id": "187899923"},  # start searching for segments at A
            "B": {"osm_node_id": "187942339"},
        },
        {
            "link": [{"name": ["I 35E"]}],
            "A": {"osm_node_id": "961117623"},  # start searching for segments at A
            "B": {"osm_node_id": "2564047368"},
        },
        {
            "link": [
                {"name": ["6th", "Sixth", "sixth"]},
                {"model_link_id": [2846, 2918]},
                {"lanes": [1, 2]},
            ]
        },
    ],
)
@pytest.mark.roadway
@pytest.mark.travis
def test_select_roadway_features(request, selection):
    print("\n--Starting:", request.node.name)
    net = _read_stpaul_net()
    print("--->", selection)
    net.select_roadway_features(selection)

    print("building a selection key")
    sel_key = net.build_selection_key(selection)

    print("Features selected:", len(net.selections[sel_key]["selected_links"]))
    selected_link_indices = net.selections[sel_key]["selected_links"].index.tolist()
    if "answer" in selection.keys():
        print("Comparing answer")
        selected_nodes = [str(selection["A"]["osm_node_id"])] + net.links_df.loc[
            selected_link_indices, "v"
        ].tolist()
        # print("Nodes selected: ",selected_nodes)
        # print("Expected Answer: ", sel["answer"])
        assert set(selected_nodes) == set(selection["answer"])

    print("--Finished:", request.node.name)


@pytest.mark.roadway
@pytest.mark.travis
def test_select_roadway_features_from_projectcard(request):
    print("\n--Starting:", request.node.name)
    net = _read_stpaul_net()
    print("Reading project card ...")
    project_card_name = "3_multiple_roadway_attribute_change.yml"

    project_card_path = os.path.join(STPAUL_DIR, "project_cards", project_card_name)
    project_card = ProjectCard.read(project_card_path)

    print("Selecting roadway features ...")
    sel = project_card.facility
    selected_link_indices = net.select_roadway_features(sel)
    print("Features selected:", len(selected_link_indices))

    print("--Finished:", request.node.name)


@pytest.mark.parametrize(
    "apply_feature_change_project_card",
    [
        "1_simple_roadway_attribute_change.yml",
        "2_multiple_roadway.yml",
        "3_multiple_roadway_attribute_change.yml",
    ],
)
@pytest.mark.roadway
@pytest.mark.travis
def test_apply_roadway_feature_change(request, apply_feature_change_project_card):
    print("\n--Starting:", request.node.name)
    my_net = _read_stpaul_net()
    print("Reading project card", apply_feature_change_project_card, "...")
    project_card_path = os.path.join(
        STPAUL_DIR, "project_cards", apply_feature_change_project_card
    )
    project_card = ProjectCard.read(project_card_path)

    print("Selecting roadway features ...")
    selected_link_indices = my_net.select_roadway_features(project_card.facility)

    attributes_to_update = [p["property"] for p in project_card.properties]
    orig_links = my_net.links_df.loc[selected_link_indices, attributes_to_update]
    print("Original Links:\n", orig_links)

    my_net.apply_roadway_feature_change(
        my_net.select_roadway_features(project_card.facility), project_card.properties
    )

    rev_links = my_net.links_df.loc[selected_link_indices, attributes_to_update]
    print("Revised Links:\n", rev_links)

    print("--Finished:", request.node.name)


@pytest.mark.roadway
@pytest.mark.travis
def test_add_managed_lane(request):
    print("\n--Starting:", request.node.name)
    net = _read_stpaul_net()
    print("Reading project card ...")
    project_card_name = "5_managed_lane.yml"
    project_card_path = os.path.join(STPAUL_DIR, "project_cards", project_card_name)
    project_card = ProjectCard.read(project_card_path)

    print("Selecting roadway features ...")
    selected_link_indices = net.select_roadway_features(project_card.facility)

    attributes_to_update = [p["property"] for p in project_card.properties]
    orig_links = net.links_df.loc[selected_link_indices, attributes_to_update]
    print("Original Links:\n", orig_links)

    net.apply_managed_lane_feature_change(
        net.select_roadway_features(project_card.facility), project_card.properties
    )

    rev_links = net.links_df.loc[selected_link_indices, attributes_to_update]
    print("Revised Links:\n", rev_links)

    net.write(filename="test_ml", path=SCRATCH_DIR)

    print("--Finished:", request.node.name)


@pytest.mark.roadway
@pytest.mark.travis
def test_add_adhoc_field(request):
    """
    Makes sure new fields can be added in the API and be saved and read in again.
    """
    print("\n--Starting:", request.node.name)
    net = _read_stpaul_net()
    net.links_df["my_ad_hoc_field"] = 22.5

    print("Network with field...\n ", net.links_df["my_ad_hoc_field"][0:5])

    assert net.links_df["my_ad_hoc_field"][0] == 22.5


@pytest.mark.roadway
@pytest.mark.travis
def test_add_adhoc_field_from_card(request):
    """
    Makes sure new fields can be added from a project card and that
    they will be the right type.
    """
    print("\n--Starting:", request.node.name)
    net = _read_stpaul_net()
    project_card_name = "new_fields_project_card.yml"

    print("Reading project card", project_card_name, "...")
    project_card_path = os.path.join(STPAUL_DIR, "project_cards", project_card_name)
    project_card = ProjectCard.read(project_card_path)

    print("Selecting roadway features ...")
    selected_link_indices = net.select_roadway_features(project_card.facility)

    attributes_to_update = [p["property"] for p in project_card.properties]

    net.apply_roadway_feature_change(
        net.select_roadway_features(project_card.facility), project_card.properties
    )

    rev_links = net.links_df.loc[selected_link_indices, attributes_to_update]
    rev_types = [(a, net.links_df[a].dtypes) for a in attributes_to_update]
    # rev_types = net.links_df[[attributes_to_update]].dtypes
    print("Revised Links:\n", rev_links, "\nNew Property Types:\n", rev_types)

    assert net.links_df.loc[selected_link_indices[0], "my_ad_hoc_field_float"] == 1.1
    assert net.links_df.loc[selected_link_indices[0], "my_ad_hoc_field_integer"] == 2
    assert (
        net.links_df.loc[selected_link_indices[0], "my_ad_hoc_field_string"] == "three"
    )
    print("--Finished:", request.node.name)


@pytest.mark.roadway
@pytest.mark.travis
def test_bad_properties_statements(request):
    """
    Makes sure new fields can be added from a project card and that
    they will be the right type.
    """

    print("\n--Starting:", request.node.name)
    net = _read_stpaul_net()
    ok_properties_change = [{"property": "lanes", "change": 1}]
    bad_properties_change = [{"property": "my_random_var", "change": 1}]
    bad_properties_existing = [{"property": "my_random_var", "existing": 1}]

    with pytest.raises(ValueError):
        net.validate_properties(bad_properties_change)

    with pytest.raises(ValueError):
        net.validate_properties(ok_properties_change, require_existing_for_change=True)

    with pytest.raises(ValueError):
        net.validate_properties(bad_properties_existing, ignore_existing=False)

    print("--Finished:", request.node.name)


@pytest.mark.travis
@pytest.mark.roadway
def test_add_delete_roadway_project_card(request):
    print("\n--Starting:", request.node.name)

    print("Reading network ...")
    net = _read_stpaul_net()
    project_cards_list = [
        "10_simple_roadway_add_change.yml",
        "11_multiple_roadway_add_and_delete_change.yml",
    ]

    for card_name in project_cards_list:
        print("Applying project card - ", card_name, "...")
        project_card_path = os.path.join(STPAUL_DIR, "project_cards", card_name)
        project_card = ProjectCard.read(project_card_path, validate=False)

        orig_links_count = len(net.links_df)
        orig_nodes_count = len(net.nodes_df)

        missing_nodes = []
        missing_links = []

        project_card_dictionary = project_card.__dict__

        def _get_missing_nodes(project_card_dictionary):
            if project_card_dictionary[
                "category"
            ].lower() == "roadway deletion" and project_card_dictionary.get("nodes"):
                for key, val in project_card_dictionary["nodes"].items():
                    return [v for v in val if v not in net.nodes_df[key].tolist()]

        def _get_missing_links(project_card_dictionary):
            if project_card_dictionary[
                "category"
            ].lower() == "roadway deletion" and project_card_dictionary.get("links"):
                for key, val in project_card_dictionary["links"].items():
                    return [v for v in val if v not in net.links_df[key].tolist()]

        # count nodes that are in original network that should be deleted
        if not project_card_dictionary.get("changes"):
            m_n = _get_missing_nodes(project_card_dictionary)
            if m_n:
                missing_nodes += m_n

            m_l = _get_missing_links(project_card_dictionary)
            if m_l:
                missing_links += m_l
        else:
            for project_dictionary in project_card_dictionary["changes"]:
                m_n = _get_missing_nodes(project_dictionary)
                if m_n:
                    missing_nodes += m_n

                m_l = _get_missing_links(project_dictionary)
                if m_l:
                    missing_links += m_l

        net.apply(project_card.__dict__)

        rev_links_count = len(net.links_df)
        rev_nodes_count = len(net.nodes_df)

        def _count_add_or_delete_features(project_card_dictionary):
            _links_added = 0
            _nodes_added = 0
            _links_deleted = 0
            _nodes_deleted = 0

            if project_card_dictionary["category"].lower() == "add new roadway":
                if project_card_dictionary.get("links") is not None:
                    _links_added = len(project_card_dictionary["links"])
                if project_card_dictionary.get("nodes") is not None:
                    _nodes_added = len(project_card_dictionary["nodes"])

            if project_card_dictionary["category"].lower() == "roadway deletion":
                if project_card_dictionary.get("links") is not None:
                    print("links", project_card_dictionary["links"])
                    _links_deleted = sum(
                        len(project_card_dictionary["links"][key])
                        for key in project_card_dictionary["links"]
                    )
                if project_card_dictionary.get("nodes"):
                    print("nodes", project_card_dictionary["nodes"])
                    _nodes_deleted = sum(
                        len(project_card_dictionary["nodes"][key])
                        for key in project_card_dictionary["nodes"]
                    )
                    print("nodes_deleted:", _nodes_deleted)
                    print("NODES:", project_card_dictionary["nodes"])

            return {
                "links_added": _links_added,
                "nodes_added": _nodes_added,
                "links_deleted": _links_deleted,
                "nodes_deleted": _nodes_deleted,
            }

        links_added = 0
        links_deleted = 0
        nodes_added = 0
        nodes_deleted = 0

        if not project_card_dictionary.get("changes"):
            count_info = _count_add_or_delete_features(project_card_dictionary)
            links_added = count_info["links_added"]
            links_deleted = count_info["links_deleted"]
            nodes_added = count_info["nodes_added"]
            nodes_deleted = count_info["nodes_deleted"]
        else:
            for project_dictionary in project_card_dictionary["changes"]:
                count_info = _count_add_or_delete_features(project_dictionary)
                links_added += count_info["links_added"]
                links_deleted += count_info["links_deleted"]
                nodes_added += count_info["nodes_added"]
                nodes_deleted += count_info["nodes_deleted"]

        num_missing_nodes = len(set(missing_nodes))
        print("MISSING NODES:", num_missing_nodes)

        num_missing_links = len(set(missing_links))
        print("MISSING LINKS:", num_missing_links)

        net_links_network = rev_links_count - orig_links_count
        net_links_project_card = links_added - links_deleted + num_missing_links

        net_nodes_network = rev_nodes_count - orig_nodes_count
        net_nodes_project_card = nodes_added - nodes_deleted + num_missing_nodes
        assert net_links_network == net_links_project_card
        assert net_nodes_network == net_nodes_project_card

    print("--Finished:", request.node.name)


@pytest.mark.roadway
@pytest.mark.travis
def test_export_network_to_csv(request):
    print("\n--Starting:", request.node.name)
    net = _read_stpaul_net()
    net.links_df.to_csv(os.path.join(SCRATCH_DIR, "links_df.csv"), index=False)
    net.nodes_df.to_csv(os.path.join(SCRATCH_DIR, "nodes_df.csv"), index=False)


variable_queries = [
    {"v": "lanes", "category": None, "time_period": ["7:00", "9:00"]},
    {"v": "ML_price", "category": "sov", "time_period": ["7:00", "9:00"]},
    {"v": "ML_price", "category": ["hov3", "hov2"], "time_period": ["7:00", "9:00"]},
]


@pytest.mark.parametrize("variable_query", variable_queries)
@pytest.mark.roadway
def test_query_roadway_property_by_time_group(request, variable_query):
    print("\n--Starting:", request.node.name)
    net = _read_stpaul_net()
    print("Applying project card...")
    project_card_path = os.path.join(STPAUL_DIR, "project_cards", "5_managed_lane.yml")
    project_card = ProjectCard.read(project_card_path, validate=False)
    net.apply_managed_lane_feature_change(
        net.select_roadway_features(project_card.facility), project_card.properties
    )
    print("Querying Attribute...")
    print("QUERY:\n", variable_query)
    v_series = net.get_property_by_time_period_and_group(
        variable_query["v"],
        category=variable_query["category"],
        time_period=variable_query["time_period"],
    )
    selected_link_indices = net.select_roadway_features(project_card.facility)

    print("CALCULATED:\n", v_series.loc[selected_link_indices])
    print("ORIGINAL:\n", net.links_df.loc[selected_link_indices, variable_query["v"]])

    ## todo make test make sure the values are correct.


@pytest.mark.highway
@pytest.mark.travis
def test_write_model_net(request):
    print("\n--Starting:", request.node.name)

    print("Reading network ...")

    net = RoadwayNetwork.read(
        link_file=STPAUL_LINK_FILE,
        node_file=STPAUL_NODE_FILE,
        shape_file=STPAUL_SHAPE_FILE,
        fast=True,
    )

    print("Reading project card ...")
    project_card_name = "5_managed_lane.yml"
    project_card_path = os.path.join(STPAUL_DIR, "project_cards", project_card_name)
    project_card = ProjectCard.read(project_card_path)

    net.apply(project_card.__dict__)
    net.links_df.to_csv(os.path.join(SCRATCH_DIR, "in_ml_links.csv"), index=False)
    net.nodes_df.to_csv(os.path.join(SCRATCH_DIR, "in_ml_nodes.csv"), index=False)
    net.shapes_df.to_csv(os.path.join(SCRATCH_DIR, "in_ml_shape.csv"), index=False)

    ml_net = net.create_managed_lane_network(in_place=False)
    ml_net.links_df.to_csv(os.path.join(SCRATCH_DIR, "out_ml_links.csv"), index=False)
    ml_net.nodes_df.to_csv(os.path.join(SCRATCH_DIR, "out_ml_nodes.csv"), index=False)
    ml_net.shapes_df.to_csv(os.path.join(SCRATCH_DIR, "out_ml_shape.csv"), index=False)

    print("--Finished:", request.node.name)


@pytest.mark.offset
@pytest.mark.travis
def test_lat_lon_offset(request):
    print("\n--Starting:", request.node.name)

    in_lat_lon = [-93.0903549, 44.961085]
    print(in_lat_lon)

    new_lat_lon = offset_lat_lon(in_lat_lon)
    print(new_lat_lon)

    print("--Finished:", request.node.name)


@pytest.mark.get_dist
@pytest.mark.travis
def test_get_distance_bw_lat_lon(request):
    print("\n--Starting:", request.node.name)

    start = [-93.0889873, 44.966861]
    end = [-93.08844310000001, 44.9717832]
    dist = haversine_distance(start, end)
    print(dist)

    print("--Finished:", request.node.name)


@pytest.mark.highway
@pytest.mark.travis
def test_network_connectivity(request):
    print("\n--Starting:", request.node.name)

    print("Reading network ...")

    net = RoadwayNetwork.read(
        link_file=STPAUL_LINK_FILE,
        node_file=STPAUL_NODE_FILE,
        shape_file=STPAUL_SHAPE_FILE,
        fast=True,
    )
    print("Checking network connectivity ...")
    print("Drive Network Connected:", net.is_network_connected(mode="drive"))
    print("--Finished:", request.node.name)

@pytest.mark.highway
@pytest.mark.travis
def test_get_modal_network(request):
    print("\n--Starting:", request.node.name)

    mode = "transit"
    print("Reading network. Mode: {} ...".format(mode))

    net = RoadwayNetwork.read(
        link_file=STPAUL_LINK_FILE,
        node_file=STPAUL_NODE_FILE,
        shape_file=STPAUL_SHAPE_FILE,
        fast=True,
    )
    _links_df, _nodes_df = RoadwayNetwork.get_modal_links_nodes(
        net.links_df, net.nodes_df, mode=mode,
    )
    mode_variable = RoadwayNetwork.MODES_TO_NETWORK_LINK_VARIABLES[mode]
    non_transit_links =_links_df[_links_df[mode_variable] != 1]
    assert non_transit_links.shape[0] == 0

@pytest.mark.highway
@pytest.mark.travis
@pytest.mark.menow
def test_network_connectivity_ignore_single_nodes(request):
    print("\n--Starting:", request.node.name)

    print("Reading network ...")

    net = RoadwayNetwork.read(
        link_file=STPAUL_LINK_FILE,
        node_file=STPAUL_NODE_FILE,
        shape_file=STPAUL_SHAPE_FILE,
        fast=True,
    )
    print("Assessing network connectivity for walk...")
    _, disconnected_nodes = net.assess_connectivity(mode="walk", ignore_end_nodes=True)
    print("{} Disconnected Subnetworks:".format(len(disconnected_nodes)))
    print("-->\n{}".format("\n".join(list(map(str,disconnected_nodes)))))
    print("--Finished:", request.node.name)

@pytest.mark.roadway
@pytest.mark.travis
@pytest.mark.xfail(strict=True)
def test_add_roadway_links(request):
    print("\n--Starting:", request.node.name)
    net = _read_stpaul_net()

    print("Reading project card ...")
    # project_card_name = "10_simple_roadway_add_change.yml"
    project_card_name = "10a_incorrect_roadway_add_change.yml"

    project_card_path = os.path.join(STPAUL_DIR, "project_cards", project_card_name)
    project_card = ProjectCard.read(project_card_path)

    project_card_dictionary = project_card.__dict__

    net.add_new_roadway_feature_change(
        project_card_dictionary.get("links"), project_card_dictionary.get("nodes")
    )

    print("--Finished:", request.node.name)



__version__ = "0.0.0"

import sys

from .Logger import WranglerLogger, setupLogging
from .ProjectCard import ProjectCard
from .RoadwayNetwork import RoadwayNetwork
from .TransitNetwork import TransitNetwork
from .Scenario import Scenario
from .Utils import point_df_to_geojson, link_df_to_json, make_slug
from .Utils import parse_time_spans, offset_lat_lon, haversine_distance

__all__ = [
    "WranglerLogger",
    "setupLogging",
    "ProjectCard",
    "RoadwayNetwork",
    "point_df_to_geojson",
    "TransitNetwork",
    "Scenario",
    "link_df_to_json",
    "make_slug",
    "parse_time_spans",
    "offset_lat_lon",
    "haversine_distance",
]

setupLogging(logFileName="network_wrangler.log")



import os
import json
import pytest
from network_wrangler import RoadwayNetwork
from network_wrangler import TransitNetwork
from network_wrangler import ProjectCard


"""
Run just the tests labeled transit using `pytest -v -m transit`
"""

STPAUL_DIR = os.path.join(os.getcwd(), "examples", "stpaul")
SCRATCH_DIR = os.path.join(os.getcwd(), "scratch")


@pytest.mark.transit_with_roadnet
@pytest.mark.travis
def test_set_roadnet(request):
    print("\n--Starting:", request.node.name)

    road_net = RoadwayNetwork.read(
        link_file=os.path.join(STPAUL_DIR, "link.json"),
        node_file=os.path.join(STPAUL_DIR, "node.geojson"),
        shape_file=os.path.join(STPAUL_DIR, "shape.geojson"),
        fast=True
    )
    transit_net = TransitNetwork.read(STPAUL_DIR)
    transit_net.set_roadnet(road_net)

    print("--Finished:", request.node.name)


@pytest.mark.transit_with_roadnet
@pytest.mark.travis
@pytest.mark.skip("need to update transit routing project card with new network")
def test_project_card(request):
    print("\n--Starting:", request.node.name)

    transit_net = TransitNetwork.read(STPAUL_DIR)
    project_card_path = os.path.join(
        STPAUL_DIR, "project_cards", "12_transit_shape_change.yml"
    )
    project_card = ProjectCard.read(project_card_path)
    transit_net.apply_transit_feature_change(
        transit_net.select_transit_features(project_card.facility),
        project_card.properties
    )

    # Shapes
    result = transit_net.feed.shapes[transit_net.feed.shapes["shape_id"] ==
                                     "2940002"]["shape_model_node_id"].tolist()
    answer = ["45983", "126312", "126316", "46663", "46665", "150855", "11188",
              "84899", "46666", "77077", "68609", "39425", "62146", "41991", "70841",
              "45691", "69793", "45683", "45685", "7688", "45687", "100784", "100782",
              "45688", "37609", "19077", "38696", "91685", "38698", "138152", "91825",
              "71086", "133190", "133187", "133188", "133183", "133179", "133177",
              "44298", "67125", "68417", "72311", "75802", "46083", "75783", "71964",
              "71456", "44190", "61464", "75786", "74898", "73817", "51814", "75787",
              "75122", "75788", "123002", "123003", "39803", "75789", "70300", "68072",
              "68039", "68041", "68043", "68044", "40878", "32484", "32474", "32485",
              "32499", "32532", "32544", "32479", "32491", "32502", "10658", "10655",
              "10651", "10652", "10649", "111619", "10643", "10642", "10640", "74056",
              "70553", "12145", "12147", "56079", "12142", "79492", "76595", "51578",
              "74179", "82208", "41799", "123353", "123351", "51651", "67557", "12136",
              "12135",  "67419", "56842", "80494", "128663", "52195", "12137", "12146",
              "25816", "25809", "115180", "9910", "115159", "115154", "162573", "10409",
              "164501", "170453", "158052", "145473", "25658",  "91126", "90985", "90993",
              "145453", "112931", "157281", "163689",  "163713", "117979", "116895",
              "138571", "87210", "97813", "165113",  "133948", "101458", "105285",
              "100603", "165025", "127797", "7753",  "127801", "28889", "28890", "127804",
              "7754", "127802", "48699",  "125802", "125237", "96386", "96366", "96367",
              "145532", "94231", "6578", "7278", "9831", "10910", "10242", "112268",
              "110085", "114928", "114941", "11694", "11654", "11705", "154299", "11708",
              "11710",  "11704", "11674", "11668", "11677", "9483", "161536", "88542",
              "22645", "19792", "126666", "170579", "145513", "155886", "101864",
              "100562",  "51955", "23270", "23263", "23271", "139930"]
    assert result == answer

    # Stops
    result = transit_net.feed.stop_times[transit_net.feed.stop_times["trip_id"] ==
                                         "14944022-JUN19-MVS-BUS-Weekday-01"]["stop_id"].tolist()
    result_tail = result[-5:]
    answer_tail = ["16842", "16837", "16836", "16833", "17040"]
    assert result_tail == answer_tail

    print("--Finished:", request.node.name)


@pytest.mark.transit_with_roadnet
@pytest.mark.travis
@pytest.mark.skip("need to allow for creating new stops if they don't already exist in stops.txt")
def test_wo_existing(request):
    print("\n--Starting:", request.node.name)

    transit_net = TransitNetwork.read(STPAUL_DIR)

    # A new node ID (not in stops.txt) should fail right now
    with pytest.raises(Exception):
        transit_net.apply_transit_feature_change(
            trip_ids=transit_net.select_transit_features(
                {"trip_id": ["14944022-JUN19-MVS-BUS-Weekday-01"]}
            ),
            properties=[
                {
                    "property": "routing",
                    "set": [1]
                }
            ]
        )

    transit_net.apply_transit_feature_change(
        trip_ids=transit_net.select_transit_features(
            {"trip_id": ["14986385-JUN19-MVS-BUS-Weekday-01"]}
        ),
        properties=[
            {
                "property": "routing",
                "set": [75318]
            }
        ]
    )

    # Shapes
    result = transit_net.feed.shapes[
        transit_net.feed.shapes["shape_id"] == "210005"
    ]["shape_model_node_id"].tolist()
    answer = ["1"]
    assert result == answer

    # Stops
    result = transit_net.feed.stop_times[
        transit_net.feed.stop_times["trip_id"] == "14986385-JUN19-MVS-BUS-Weekday-01"
    ]["stop_id"].tolist()
    answer = ["2609"]  # first matching stop_id in stops.txt
    assert result == answer

    print("--Finished:", request.node.name)


@pytest.mark.transit_with_roadnet
@pytest.mark.travis
@pytest.mark.skip("need to update trips and nodes")
def test_select_transit_features_by_nodes(request):
    print("\n--Starting:", request.node.name)

    transit_net = TransitNetwork.read(STPAUL_DIR)

    # Any nodes
    trip_ids = transit_net.select_transit_features_by_nodes(
        node_ids=["29636", "29666"]
    )
    assert set(trip_ids) == set([
        "14940701-JUN19-MVS-BUS-Weekday-01",
        "14942968-JUN19-MVS-BUS-Weekday-01"
    ])

    # All nodes
    trip_ids = transit_net.select_transit_features_by_nodes(
        node_ids=["29636", "29666"], require_all=True
    )
    assert set(trip_ids) == set([
        "14940701-JUN19-MVS-BUS-Weekday-01"
    ])

    print("--Finished:", request.node.name)


@pytest.mark.transit_with_roadnet
@pytest.mark.travis
@pytest.mark.skip("need to update trips and nodes")
def test_select_transit_features_by_nodes(request):
    print("\n--Starting:", request.node.name)

    transit_net = TransitNetwork.read(STPAUL_DIR)

    # Any nodes
    trip_ids = transit_net.select_transit_features_by_nodes(
        node_ids=["29636", "29666"]
    )
    assert set(trip_ids) == set([
        "14940701-JUN19-MVS-BUS-Weekday-01",
        "14942968-JUN19-MVS-BUS-Weekday-01"
    ])

    # All nodes
    trip_ids = transit_net.select_transit_features_by_nodes(
        node_ids=["29636", "29666"], require_all=True
    )
    assert set(trip_ids) == set([
        "14940701-JUN19-MVS-BUS-Weekday-01"
    ])

    print("--Finished:", request.node.name)



import os

import json, yaml
import pytest
from jsonschema import validate

SCHEMA_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
    "network_wrangler",
    "schemas",
)
SMALL_EX_DIR = os.path.join(os.getcwd(), "examples", "single")
STPAUL_EX_DIR = os.path.join(os.getcwd(), "examples", "stpaul")

STPAUL_PC_DIR = os.path.join(os.getcwd(), "examples", "stpaul", "project_cards")


@pytest.mark.schema
@pytest.mark.roadschema
@pytest.mark.skip(reason="need to work on this")
def test_roadway_link_schema():
    schema_filename = os.path.join(SCHEMA_DIR, "roadway_network_link.json")
    link_file = os.path.join(SMALL_EX_DIR, "link.json")

    with open(schema_filename) as schema_json_file:
        schema = json.load(schema_json_file)

    with open(link_file, "r") as links:
        link_json = yaml.safe_load(links)

    validate(link_json, schema)


@pytest.mark.schema
@pytest.mark.skip(reason="need to work on this")
def test_project_card_schema():
    schema_filename = os.path.join(SCHEMA_DIR, "project_card.json")
    card_file = os.path.join(STPAUL_PC_DIR, "1_simple_roadway_attribute_change.yml")

    with open(schema_filename) as schema_json_file:
        schema = json.load(schema_json_file)

    with open(card_file, "r") as card:
        card_json = yaml.safe_load(card)

    validate(card_json, schema)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import os
import sys
import copy

import yaml
import pandas as pd
import geojson
import geopandas as gpd
import json
import networkx as nx
import numpy as np

from pandas.core.frame import DataFrame
from geopandas.geodataframe import GeoDataFrame

from jsonschema import validate
from jsonschema.exceptions import ValidationError
from jsonschema.exceptions import SchemaError

import folium
import osmnx as ox
from shapely.geometry import Point, LineString
from random import randint

from .Logger import WranglerLogger
from .Utils import point_df_to_geojson, link_df_to_json, parse_time_spans
from .Utils import offset_lat_lon, haversine_distance
from .ProjectCard import ProjectCard


class RoadwayNetwork(object):
    """
    Representation of a Roadway Network.

    Attributes:
        nodes_df (GeoDataFrame): node data

        links_df (GeoDataFrame): link data, including start and end
            nodes and associated shape

        shapes_df (GeoDataFrame): detailed shape data

        selections (dict): dictionary storing selections in case they are made repeatedly

        CRS (str): coordinate reference system in PROJ4 format.
            See https://proj.org/operations/projections/index.html#

        ESPG (int): integer representing coordinate system https://epsg.io/

        NODE_FOREIGN_KEY (str): column in `nodes_df` associated with the
            LINK_FOREIGN_KEY

        LINK_FOREIGN_KEY (list(str)): list of columns in `link_df` that
            represent the NODE_FOREIGN_KEY

        UNIQUE_LINK_KEY (str): column that is a unique key for links

        UNIQUE_NODE_KEY (str): column that is a unique key for nodes

        UNIQUE_MODEL_LINK_IDENTIFIERS (list(str)): list of all unique
            identifiers for links, including the UNIQUE_LINK_KEY
        UNIQUE_NODE_IDENTIFIERS (list(str)): list of all unique identifiers
            for nodes, including the UNIQUE_NODE_KEY

        SELECTION_REQUIRES (list(str))): required attributes in the selection
            if a unique identifier is not used

        SEARCH_BREADTH (int): initial number of links from name-based
            selection that are traveresed before trying another shortest
            path when searching for paths between A and B node

        MAX_SEARCH_BREADTH (int): maximum number of links traversed between
            links that match the searched name when searching for paths
            between A and B node

        SP_WEIGHT_FACTOR (Union(int, float)): penalty assigned for each
            degree of distance between a link and a link with the searched-for
            name when searching for paths between A and B node

        MANAGED_LANES_TO_NODE_ID_SCALAR (int): scalar value added to
            the general purpose lanes' `model_node_id` when creating
            an associated node for a parallel managed lane

        MANAGED_LANES_TO_LINK_ID_SCALAR (int): scalar value added to
            the general purpose lanes' `model_link_id` when creating
            an associated link for a parallel managed lane

        MANAGED_LANES_REQUIRED_ATTRIBUTES (list(str)): list of attributes
            that must be provided in managed lanes

        KEEP_SAME_ATTRIBUTES_ML_AND_GP (list(str)): list of attributes
            to copy from a general purpose lane to managed lane
    """

    CRS = "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs"
    EPSG = 4326

    NODE_FOREIGN_KEY = "model_node_id"
    LINK_FOREIGN_KEY = ["A", "B"]

    SEARCH_BREADTH = 5
    MAX_SEARCH_BREADTH = 10
    SP_WEIGHT_FACTOR = 100
    MANAGED_LANES_NODE_ID_SCALAR = 500000
    MANAGED_LANES_LINK_ID_SCALAR = 1000000

    SELECTION_REQUIRES = ["link"]

    UNIQUE_LINK_KEY = "model_link_id"
    UNIQUE_NODE_KEY = "model_node_id"
    UNIQUE_MODEL_LINK_IDENTIFIERS = ["model_link_id"]
    UNIQUE_NODE_IDENTIFIERS = ["model_node_id"]

    MANAGED_LANES_REQUIRED_ATTRIBUTES = [
        "A",
        "B",
        "model_link_id",
        "locationReferences",
    ]

    KEEP_SAME_ATTRIBUTES_ML_AND_GP = [
        "distance",
        "bike_access",
        "drive_access",
        "transit_access",
        "walk_access",
        "maxspeed",
        "name",
        "oneway",
        "ref",
        "roadway",
        "length",
    ]

    MANAGED_LANES_SCALAR = 500000

    MODES_TO_NETWORK_LINK_VARIABLES = {
        "drive": "drive_access",
        "transit": "transit_access",
        "walk": "walk_access",
        "bike": "bike_access",
    }

    MODES_TO_NETWORK_NODE_VARIABLES = {
        "drive": "drive_node",
        "transit": "transit_node",
        "walk": "walk_node",
        "bike": "bike_node",
    }

    def __init__(self, nodes: GeoDataFrame, links: GeoDataFrame, shapes: GeoDataFrame):
        """
        Constructor
        """

        if not RoadwayNetwork.validate_object_types(nodes, links, shapes):
            sys.exit("RoadwayNetwork: Invalid constructor data type")

        self.nodes_df = nodes
        self.links_df = links
        self.shapes_df = shapes

        self.link_file = None
        self.node_file = None
        self.shape_file = None

        # Add non-required fields if they aren't there.
        # for field, default_value in RoadwayNetwork.OPTIONAL_FIELDS:
        #    if field not in self.links_df.columns:
        #        self.links_df[field] = default_value
        if not self.validate_uniqueness():
            raise ValueError("IDs in network not unique")
        self.selections = {}

    @staticmethod
    def read(
        link_file: str, node_file: str, shape_file: str, fast: bool = True
    ) -> RoadwayNetwork:
        ##TODO turn off fast=True as default
        """
        Reads a network from the roadway network standard
        Validates that it conforms to the schema

        args:
            link_file: full path to the link file
            node_file: full path to the node file
            shape_file: full path to the shape file
            fast: boolean that will skip validation to speed up read time

        Returns: a RoadwayNetwork instance
        """

        WranglerLogger.info(
            "Reading from following files:\n-{}\n-{}\n-{}.".format(
                link_file, node_file, shape_file
            )
        )

        """
        Validate Input
        """

        if not os.path.exists(link_file):
            msg = "Link file doesn't exist at: {}".format(link_file)
            WranglerLogger.error(msg)
            raise ValueError(msg)

        if not os.path.exists(node_file):
            msg = "Link file doesn't exist at: {}".format(link_file)
            WranglerLogger.error(msg)
            raise ValueError(msg)

        if not os.path.exists(shape_file):
            msg = "Link file doesn't exist at: {}".format(link_file)
            WranglerLogger.error(msg)
            raise ValueError(msg)

        if not fast:
            if not (
                RoadwayNetwork.validate_node_schema(node_file)
                and RoadwayNetwork.validate_link_schema(link_file)
                and RoadwayNetwork.validate_shape_schema(shape_file)
            ):

                sys.exit("RoadwayNetwork: Data doesn't conform to schema")

        with open(link_file) as f:
            link_json = json.load(f)

        link_properties = pd.DataFrame(link_json)
        link_geometries = [
            LineString(
                [
                    g["locationReferences"][0]["point"],
                    g["locationReferences"][1]["point"],
                ]
            )
            for g in link_json
        ]
        links_df = gpd.GeoDataFrame(link_properties, geometry=link_geometries)
        links_df.crs = RoadwayNetwork.CRS

        shapes_df = gpd.read_file(shape_file)
        shapes_df.dropna(subset=["geometry", "id"], inplace=True)
        shapes_df.crs = RoadwayNetwork.CRS

        # geopandas uses fiona OGR drivers, which doesn't let you have
        # a list as a property type. Therefore, must read in node_properties
        # separately in a vanilla dataframe and then convert to geopandas

        with open(node_file) as f:
            node_geojson = json.load(f)

        node_properties = pd.DataFrame(
            [g["properties"] for g in node_geojson["features"]]
        )
        node_geometries = [
            Point(g["geometry"]["coordinates"]) for g in node_geojson["features"]
        ]

        nodes_df = gpd.GeoDataFrame(node_properties, geometry=node_geometries)

        nodes_df.gdf_name = "network_nodes"

        # set a copy of the  foreign key to be the index so that the
        # variable itself remains queryiable
        nodes_df[RoadwayNetwork.NODE_FOREIGN_KEY + "_idx"] = nodes_df[
            RoadwayNetwork.NODE_FOREIGN_KEY
        ]
        nodes_df.set_index(RoadwayNetwork.NODE_FOREIGN_KEY + "_idx", inplace=True)

        nodes_df.crs = RoadwayNetwork.CRS
        nodes_df["x"] = nodes_df["geometry"].apply(lambda g: g.x)
        nodes_df["y"] = nodes_df["geometry"].apply(lambda g: g.y)
        # todo: flatten json

        WranglerLogger.info("Read %s links from %s" % (links_df.size, link_file))
        WranglerLogger.info("Read %s nodes from %s" % (nodes_df.size, node_file))
        WranglerLogger.info("Read %s shapes from %s" % (shapes_df.size, shape_file))

        roadway_network = RoadwayNetwork(
            nodes=nodes_df, links=links_df, shapes=shapes_df
        )

        roadway_network.link_file = link_file
        roadway_network.node_file = node_file
        roadway_network.shape_file = shape_file

        return roadway_network

    def write(self, path: str = ".", filename: str = None) -> None:
        """
        Writes a network in the roadway network standard

        args:
            path: the path were the output will be saved
            filename: the name prefix of the roadway files that will be generated
        """

        if not os.path.exists(path):
            WranglerLogger.debug("\nPath [%s] doesn't exist; creating." % path)
            os.mkdir(path)

        if filename:
            links_file = os.path.join(path, filename + "_" + "link.json")
            nodes_file = os.path.join(path, filename + "_" + "node.geojson")
            shapes_file = os.path.join(path, filename + "_" + "shape.geojson")
        else:
            links_file = os.path.join(path, "link.json")
            nodes_file = os.path.join(path, "node.geojson")
            shapes_file = os.path.join(path, "shape.geojson")

        link_property_columns = self.links_df.columns.values.tolist()
        link_property_columns.remove("geometry")
        links_json = link_df_to_json(self.links_df, link_property_columns)
        with open(links_file, "w") as f:
            json.dump(links_json, f)

        # geopandas wont let you write to geojson because
        # it uses fiona, which doesn't accept a list as one of the properties
        # so need to convert the df to geojson manually first
        property_columns = self.nodes_df.columns.values.tolist()
        property_columns.remove("geometry")

        nodes_geojson = point_df_to_geojson(self.nodes_df, property_columns)

        with open(nodes_file, "w") as f:
            json.dump(nodes_geojson, f)

        self.shapes_df.to_file(shapes_file, driver="GeoJSON")

    def validate_uniqueness(self) -> Bool:
        """
        Confirms that the unique identifiers are met.
        """
        valid = True
        for c in RoadwayNetwork.UNIQUE_MODEL_LINK_IDENTIFIERS:
            if c not in self.links_df.columns:
                valid = False
                msg = "Network doesn't contain unique link identifier: {}".format(c)
                WranglerLogger.error(msg)
            if not self.links_df[c].is_unique:
                valid = False
                msg = "Unique identifier {} is not unique in network links".format(c)
                WranglerLogger.error(msg)
        for c in RoadwayNetwork.LINK_FOREIGN_KEY:
            if c not in self.links_df.columns:
                valid = False
                msg = "Network doesn't contain link foreign key identifier: {}".format(
                    c
                )
                WranglerLogger.error(msg)
        link_foreign_key = self.links_df[RoadwayNetwork.LINK_FOREIGN_KEY].apply(
            lambda x: "-".join(x.map(str)), axis=1
        )
        if not link_foreign_key.is_unique:
            valid = False
            msg = "Foreign key: {} is not unique in network links".format(
                RoadwayNetwork.LINK_FOREIGN_KEY
            )
            WranglerLogger.error(msg)
        for c in RoadwayNetwork.UNIQUE_NODE_IDENTIFIERS:
            if c not in self.nodes_df.columns:
                valid = False
                msg = "Network doesn't contain unique node identifier: {}".format(c)
                WranglerLogger.error(msg)
            if not self.nodes_df[c].is_unique:
                valid = False
                msg = "Unique identifier {} is not unique in network nodes".format(c)
                WranglerLogger.error(msg)
        if RoadwayNetwork.NODE_FOREIGN_KEY not in self.nodes_df.columns:
            valid = False
            msg = "Network doesn't contain node foreign key identifier: {}".format(
                RoadwayNetwork.NODE_FOREIGN_KEY
            )
            WranglerLogger.error(msg)
        elif not self.nodes_df[RoadwayNetwork.NODE_FOREIGN_KEY].is_unique:
            valid = False
            msg = "Foreign key: {} is not unique in network nodes".format(
                RoadwayNetwork.NODE_FOREIGN_KEY
            )
            WranglerLogger.error(msg)
        return valid

    @staticmethod
    def validate_object_types(
        nodes: GeoDataFrame, links: GeoDataFrame, shapes: GeoDataFrame
    ):
        """
        Determines if the roadway network is being built with the right object types.
        Does not validate schemas.

        Args:
            nodes: nodes geodataframe
            links: link geodataframe
            shapes: shape geodataframe

        Returns: boolean
        """

        errors = ""

        if not isinstance(nodes, GeoDataFrame):
            error_message = "Incompatible nodes type:{}. Must provide a GeoDataFrame.  ".format(
                type(nodes)
            )
            WranglerLogger.error(error_message)
            errors.append(error_message)
        if not isinstance(links, GeoDataFrame):
            error_message = "Incompatible links type:{}. Must provide a GeoDataFrame.  ".format(
                type(links)
            )
            WranglerLogger.error(error_message)
            errors.append(error_message)
        if not isinstance(shapes, GeoDataFrame):
            error_message = "Incompatible shapes type:{}. Must provide a GeoDataFrame.  ".format(
                type(shapes)
            )
            WranglerLogger.error(error_message)
            errors.append(error_message)

        if errors:
            return False
        return True

    @staticmethod
    def validate_node_schema(
        node_file, schema_location: str = "roadway_network_node.json"
    ):
        """
        Validate roadway network data node schema and output a boolean
        """
        if not os.path.exists(schema_location):
            base_path = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "schemas"
            )
            schema_location = os.path.join(base_path, schema_location)

        with open(schema_location) as schema_json_file:
            schema = json.load(schema_json_file)

        with open(node_file) as node_json_file:
            json_data = json.load(node_json_file)

        try:
            validate(json_data, schema)
            return True

        except ValidationError as exc:
            WranglerLogger.error("Failed Node schema validation: Validation Error")
            WranglerLogger.error("Node File Loc:{}".format(node_file))
            WranglerLogger.error("Node Schema Loc:{}".format(schema_location))
            WranglerLogger.error(exc.message)

        except SchemaError as exc:
            WranglerLogger.error("Invalid Node Schema")
            WranglerLogger.error("Node Schema Loc:{}".format(schema_location))
            WranglerLogger.error(json.dumps(exc.message, indent=2))

        return False

    @staticmethod
    def validate_link_schema(
        link_file, schema_location: str = "roadway_network_link.json"
    ):
        """
        Validate roadway network data link schema and output a boolean
        """

        if not os.path.exists(schema_location):
            base_path = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "schemas"
            )
            schema_location = os.path.join(base_path, schema_location)

        with open(schema_location) as schema_json_file:
            schema = json.load(schema_json_file)

        with open(link_file) as link_json_file:
            json_data = json.load(link_json_file)

        try:
            validate(json_data, schema)
            return True

        except ValidationError as exc:
            WranglerLogger.error("Failed Link schema validation: Validation Error")
            WranglerLogger.error("Link File Loc:{}".format(link_file))
            WranglerLogger.error("Path:{}".format(exc.path))
            WranglerLogger.error(exc.message)

        except SchemaError as exc:
            WranglerLogger.error("Invalid Link Schema")
            WranglerLogger.error("Link Schema Loc: {}".format(schema_location))
            WranglerLogger.error(json.dumps(exc.message, indent=2))

        return False

    @staticmethod
    def validate_shape_schema(
        shape_file, schema_location: str = "roadway_network_shape.json"
    ):
        """
        Validate roadway network data shape schema and output a boolean
        """

        if not os.path.exists(schema_location):
            base_path = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "schemas"
            )
            schema_location = os.path.join(base_path, schema_location)

        with open(schema_location) as schema_json_file:
            schema = json.load(schema_json_file)

        with open(shape_file) as shape_json_file:
            json_data = json.load(shape_json_file)

        try:
            validate(json_data, schema)
            return True

        except ValidationError as exc:
            WranglerLogger.error("Failed Shape schema validation: Validation Error")
            WranglerLogger.error("Shape File Loc:{}".format(shape_file))
            WranglerLogger.error("Path:{}".format(exc.path))
            WranglerLogger.error(exc.message)

        except SchemaError as exc:
            WranglerLogger.error("Invalid Shape Schema")
            WranglerLogger.error("Shape Schema Loc: {}".format(schema_location))
            WranglerLogger.error(json.dumps(exc.message, indent=2))

        return False

    def validate_selection(self, selection: dict) -> Bool:
        """
        Evaluate whetther the selection dictionary contains the
        minimum required values.

        Args:
            selection: selection dictionary to be evaluated

        Returns: boolean value as to whether the selection dictonary is valid.
        """
        if not set(RoadwayNetwork.SELECTION_REQUIRES).issubset(selection):
            err_msg = "Project Card Selection requires: {}".format(
                ",".join(RoadwayNetwork.SELECTION_REQUIRES)
            )
            err_msg += ", but selection only contains: {}".format(",".join(selection))
            WranglerLogger.error(err_msg)
            raise KeyError(err_msg)

        err = []
        for l in selection["link"]:
            for k, v in l.items():
                if k not in self.links_df.columns:
                    err.append(
                        "{} specified in link selection but not an attribute in network\n".format(
                            k
                        )
                    )
        selection_keys = [k for l in selection["link"] for k, v in l.items()]
        unique_link_id = bool(
            set(RoadwayNetwork.UNIQUE_MODEL_LINK_IDENTIFIERS).intersection(
                set(selection_keys)
            )
        )

        if not unique_link_id:
            for k, v in selection["A"].items():
                if (
                    k not in self.nodes_df.columns
                    and k != RoadwayNetwork.NODE_FOREIGN_KEY
                ):
                    err.append(
                        "{} specified in A node selection but not an attribute in network\n".format(
                            k
                        )
                    )
            for k, v in selection["B"].items():
                if (
                    k not in self.nodes_df.columns
                    and k != RoadwayNetwork.NODE_FOREIGN_KEY
                ):
                    err.append(
                        "{} specified in B node selection but not an attribute in network\n".format(
                            k
                        )
                    )
        if err:
            WranglerLogger.error(
                "ERROR: Selection variables in project card not found in network"
            )
            WranglerLogger.error("\n".join(err))
            WranglerLogger.error(
                "--existing node columns:{}".format(" ".join(self.nodes_df.columns))
            )
            WranglerLogger.error(
                "--existing link columns:{}".format(" ".join(self.links_df.columns))
            )
            raise ValueError()
            return False
        else:
            return True

    def orig_dest_nodes_foreign_key(
        self, selection: dict, node_foreign_key: str = ""
    ) -> tuple:
        """
        Returns the foreign key id (whatever is used in the u and v
        variables in the links file) for the AB nodes as a tuple.

        Args:
            selection : selection dictionary with A and B keys
            node_foreign_key: variable name for whatever is used by the u and v variable
            in the links_df file.  If nothing is specified, assume whatever
            default is (usually osm_node_id)

        Returns: tuple of (A_id, B_id)
        """

        if not node_foreign_key:
            node_foreign_key = RoadwayNetwork.NODE_FOREIGN_KEY
        if len(selection["A"]) > 1:
            raise ("Selection A node dictionary should be of length 1")
        if len(selection["B"]) > 1:
            raise ("Selection B node dictionary should be of length 1")

        A_node_key, A_id = next(iter(selection["A"].items()))
        B_node_key, B_id = next(iter(selection["B"].items()))

        if A_node_key != node_foreign_key:
            A_id = self.nodes_df[self.nodes_df[A_node_key] == A_id][
                node_foreign_key
            ].values[0]
        if B_node_key != node_foreign_key:
            B_id = self.nodes_df[self.nodes_df[B_node_key] == B_id][
                node_foreign_key
            ].values[0]

        return (A_id, B_id)

    @staticmethod
    def get_managed_lane_node_ids(nodes_list):
        return [x + RoadwayNetwork.MANAGED_LANES_SCALAR for x in nodes_list]

    @staticmethod
    def ox_graph(nodes_df: GeoDataFrame, links_df: GeoDataFrame):
        """
        create an osmnx-flavored network graph

        osmnx doesn't like values that are arrays, so remove the variables
        that have arrays.  osmnx also requires that certain variables
        be filled in, so do that too.

        Args:
            nodes_df : GeoDataFrame of nodes
            link_df : GeoDataFrame of links

        Returns: a networkx multidigraph
        """
        WranglerLogger.debug("starting ox_graph()")

        graph_nodes = nodes_df.copy().drop(
            ["inboundReferenceIds", "outboundReferenceIds"], axis=1
        )

        graph_nodes.gdf_name = "network_nodes"
        WranglerLogger.debug("GRAPH NODES: {}".format(graph_nodes.columns))
        graph_nodes["id"] = graph_nodes[RoadwayNetwork.NODE_FOREIGN_KEY]

        graph_links = links_df.copy().drop(
            ["osm_link_id", "locationReferences"], axis=1
        )

        # have to change this over into u,v b/c this is what osm-nx is expecting
        graph_links["u"] = graph_links[RoadwayNetwork.LINK_FOREIGN_KEY[0]]
        graph_links["v"] = graph_links[RoadwayNetwork.LINK_FOREIGN_KEY[1]]
        graph_links["id"] = graph_links[RoadwayNetwork.UNIQUE_LINK_KEY]
        graph_links["key"] = graph_links[RoadwayNetwork.UNIQUE_LINK_KEY]

        WranglerLogger.debug("starting ox.gdfs_to_graph()")
        G = ox.gdfs_to_graph(graph_nodes, graph_links)

        WranglerLogger.debug("finished ox.gdfs_to_graph()")
        return G

    @staticmethod
    def selection_has_unique_link_id(selection_dict: dict)-> bool:
        """
        Args:
            selection_dictionary: Dictionary representation of selection
                of roadway features, containing a "link" key.

        Returns: A boolean indicating if the selection dictionary contains
            a required unique link id.
        """
        selection_keys = [k for l in selection_dict["link"] for k, v in l.items()]
        return bool(
            set(RoadwayNetwork.UNIQUE_MODEL_LINK_IDENTIFIERS).intersection(
                set(selection_keys)
            )
        )

    def build_selection_key(self, selection_dict: dict)->tuple:
        """
        Selections are stored by a key combining the query and the A and B ids.
        This method combines the two for you based on the selection dictionary.

        Args:
            selection_dictonary: Selection Dictionary

        Returns: Tuple serving as the selection key.

        """
        sel_query = ProjectCard.build_link_selection_query(
            selection=selection_dict,
            unique_model_link_identifiers=RoadwayNetwork.UNIQUE_MODEL_LINK_IDENTIFIERS,
        )

        if RoadwayNetwork.selection_has_unique_link_id(selection_dict):
            return sel_query

        A_id, B_id = self.orig_dest_nodes_foreign_key(selection_dict)
        return (sel_query, A_id, B_id)

    def select_roadway_features(
        self, selection: dict, search_mode="drive", force_search=False
    ) -> GeoDataFrame:
        """
        Selects roadway features that satisfy selection criteria

        Example usage:
            net.select_roadway_features(
              selection = [ {
                #   a match condition for the from node using osm,
                #   shared streets, or model node number
                'from': {'osm_model_link_id': '1234'},
                #   a match for the to-node..
                'to': {'shstid': '4321'},
                #   a regex or match for facility condition
                #   could be # of lanes, facility type, etc.
                'facility': {'name':'Main St'},
                }, ... ])

        Args:
            selection : dictionary with keys for:
                 A - from node
                 B - to node
                 link - which includes at least a variable for `name`

        Returns: a list of node foreign IDs on shortest path
        """
        WranglerLogger.debug("validating selection")
        self.validate_selection(selection)

        # create a unique key for the selection so that we can cache it
        sel_key = self.build_selection_key(selection)
        WranglerLogger.debug("Selection Key: {}".format(sel_key))

        # if this selection has been queried before, just return the
        # previously selected links
        if sel_key in self.selections and not force_search:
            if self.selections[sel_key]["selection_found"]:
                return self.selections[sel_key]["selected_links"].index.tolist()
            else:
                msg = "Selection previously queried but no selection found"
                WranglerLogger.error(msg)
                raise Exception(msg)
        self.selections[sel_key] = {}
        self.selections[sel_key]["selection_found"] = False

        unique_model_link_identifer_in_selection = RoadwayNetwork.selection_has_unique_link_id(
            selection
        )
        if not unique_model_link_identifer_in_selection:
            A_id, B_id = self.orig_dest_nodes_foreign_key(selection)
        # identify candidate links which match the initial query
        # assign them as iteration = 0
        # subsequent iterations that didn't match the query will be
        # assigned a heigher weight in the shortest path
        WranglerLogger.debug("Building selection query")
        # build a selection query based on the selection dictionary

        sel_query = ProjectCard.build_link_selection_query(
            selection=selection,
            unique_model_link_identifiers=RoadwayNetwork.UNIQUE_MODEL_LINK_IDENTIFIERS,
            mode=RoadwayNetwork.MODES_TO_NETWORK_LINK_VARIABLES[search_mode],
        )
        WranglerLogger.debug("Selecting features:\n{}".format(sel_query))

        self.selections[sel_key]["candidate_links"] = self.links_df.query(
            sel_query, engine="python"
        )
        WranglerLogger.debug("Completed query")
        candidate_links = self.selections[sel_key][
            "candidate_links"
        ]  # b/c too long to keep that way

        candidate_links["i"] = 0

        if len(candidate_links.index) == 0 and unique_model_link_identifer_in_selection:
            msg = "No links found based on unique link identifiers.\nSelection Failed."
            WranglerLogger.error(msg)
            raise Exception(msg)

        if len(candidate_links.index) == 0:
            WranglerLogger.debug(
                "No candidate links in initial search.\nRetrying query using 'ref' instead of 'name'"
            )
            # if the query doesn't come back with something from 'name'
            # try it again with 'ref' instead
            selection_has_name_key = any("name" in d for d in selection["link"])

            if not selection_has_name_key:
                msg = "Not able to complete search using 'ref' instead of 'name' because 'name' not in search."
                WranglerLogger.error(msg)
                raise Exception(msg)

            if not "ref" in self.links_df.columns:
                msg = "Not able to complete search using 'ref' because 'ref' not in network."
                WranglerLogger.error(msg)
                raise Exception(msg)

            WranglerLogger.debug("Trying selection query replacing 'name' with 'ref'")
            sel_query = sel_query.replace("name", "ref")

            self.selections[sel_key]["candidate_links"] = self.links_df.query(
                sel_query, engine="python"
            )
            candidate_links = self.selections[sel_key]["candidate_links"]

            candidate_links["i"] = 0

            if len(candidate_links.index) == 0:
                msg = "No candidate links in search using either 'name' or 'ref' in query.\nSelection Failed."
                WranglerLogger.error(msg)
                raise Exception(msg)

        def _add_breadth(candidate_links: DataFrame, nodes: Data, links, i):
            """
            Add outbound and inbound reference IDs to candidate links
            from existing nodes

            Args:
            candidate_links : GeoDataFrame
                df with the links from the previous iteration that we
                want to add on to

            nodes : GeoDataFrame
                df of all nodes in the full network

            links : GeoDataFrame
                df of all links in the full network

            i : int
                iteration of adding breadth

            Returns:
                candidate_links : GeoDataFrame
                    updated df with one more degree of added breadth

                node_list_foreign_keys : list
                    list of foreign key ids for nodes in the updated candidate links
                    to test if the A and B nodes are in there.
            """
            WranglerLogger.debug("-Adding Breadth-")

            node_list_foreign_keys = list(
                set(
                    [
                        i
                        for fk in RoadwayNetwork.LINK_FOREIGN_KEY
                        for i in list(candidate_links[fk])
                    ]
                )
                # set(list(candidate_links["u"]) + list(candidate_links["v"]))
            )
            candidate_nodes = nodes.loc[node_list_foreign_keys]
            WranglerLogger.debug("Candidate Nodes: {}".format(len(candidate_nodes)))
            links_shstRefId_to_add = list(
                set(
                    sum(candidate_nodes["outboundReferenceIds"].tolist(), [])
                    + sum(candidate_nodes["inboundReferenceIds"].tolist(), [])
                )
                - set(candidate_links["shstReferenceId"].tolist())
                - set([""])
            )
            ##TODO make unique ID for links in the settings
            # print("Link IDs to add: {}".format(links_shstRefId_to_add))
            # print("Links: ", links_id_to_add)
            links_to_add = links[links.shstReferenceId.isin(links_shstRefId_to_add)]
            # print("Adding Links:",links_to_add)
            WranglerLogger.debug("Adding {} links.".format(links_to_add.shape[0]))
            links[links.model_link_id.isin(links_shstRefId_to_add)]["i"] = i
            candidate_links = candidate_links.append(links_to_add)
            node_list_foreign_keys = list(
                set(
                    [
                        i
                        for fk in RoadwayNetwork.LINK_FOREIGN_KEY
                        for i in list(candidate_links[fk])
                    ]
                )
                # set(list(candidate_links["u"]) + list(candidate_links["v"]))
            )

            return candidate_links, node_list_foreign_keys

        def _shortest_path():
            WranglerLogger.debug(
                "_shortest_path(): calculating shortest path from graph"
            )
            candidate_links["weight"] = 1 + (
                candidate_links["i"] * RoadwayNetwork.SP_WEIGHT_FACTOR
            )

            node_list_foreign_keys = list(
                set(
                    [
                        i
                        for fk in RoadwayNetwork.LINK_FOREIGN_KEY
                        for i in list(candidate_links[fk])
                    ]
                )
                # set(list(candidate_links["u"]) + list(candidate_links["v"]))
            )

            candidate_nodes = self.nodes_df.loc[node_list_foreign_keys]

            WranglerLogger.debug("creating network graph")
            G = RoadwayNetwork.ox_graph(candidate_nodes, candidate_links)
            self.selections[sel_key]["graph"] = G
            self.selections[sel_key]["candidate_links"] = candidate_links

            try:
                WranglerLogger.debug(
                    "Calculating NX shortest path from A_id: {} to B_id: {}".format(
                        A_id, B_id
                    )
                )
                sp_route = nx.shortest_path(G, A_id, B_id, weight="weight")
                WranglerLogger.debug("Shortest path successfully routed")

            except nx.NetworkXNoPath:
                return False

            sp_links = candidate_links[
                candidate_links["A"].isin(sp_route)
                & candidate_links["B"].isin(sp_route)
            ]

            self.selections[sel_key]["route"] = sp_route
            self.selections[sel_key]["links"] = sp_links

            return True

        if not unique_model_link_identifer_in_selection:
            # find the node ids for the candidate links
            WranglerLogger.debug("Not a unique ID selection, conduct search")
            node_list_foreign_keys = list(
                set(
                    [
                        i
                        for fk in RoadwayNetwork.LINK_FOREIGN_KEY
                        for i in list(candidate_links[fk])
                    ]
                )
                # set(list(candidate_links["u"]) + list(candidate_links["v"]))
            )
            WranglerLogger.debug("Foreign key list: {}".format(node_list_foreign_keys))
            i = 0

            max_i = RoadwayNetwork.SEARCH_BREADTH

            while (
                A_id not in node_list_foreign_keys
                and B_id not in node_list_foreign_keys
                and i <= max_i
            ):
                WranglerLogger.debug(
                    "Adding breadth, no shortest path. i: {}, Max i: {}".format(
                        i, max_i
                    )
                )
                i += 1
                candidate_links, node_list_foreign_keys = _add_breadth(
                    candidate_links, self.nodes_df, self.links_df, i
                )
            WranglerLogger.debug("calculating shortest path from graph")
            sp_found = _shortest_path()
            if not sp_found:
                WranglerLogger.info(
                    "No shortest path found with {}, trying greater breadth until SP found".format(
                        i
                    )
                )
            while not sp_found and i <= RoadwayNetwork.MAX_SEARCH_BREADTH:
                WranglerLogger.debug(
                    "Adding breadth, with shortest path iteration. i: {} Max i: {}".format(
                        i, max_i
                    )
                )
                i += 1
                candidate_links, node_list_foreign_keys = _add_breadth(
                    candidate_links, self.nodes_df, self.links_df, i
                )
                sp_found = _shortest_path()

            if sp_found:
                # reselect from the links in the shortest path, the ones with
                # the desired values....ignoring name.
                if len(selection["link"]) > 1:
                    resel_query = ProjectCard.build_link_selection_query(
                        selection=selection,
                        unique_model_link_identifiers=RoadwayNetwork.UNIQUE_MODEL_LINK_IDENTIFIERS,
                        mode=RoadwayNetwork.MODES_TO_NETWORK_LINK_VARIABLES[
                            search_mode
                        ],
                        ignore=["name"],
                    )
                    WranglerLogger.info("Reselecting features:\n{}".format(resel_query))
                    self.selections[sel_key]["selected_links"] = self.selections[
                        sel_key
                    ]["links"].query(resel_query, engine="python")
                else:
                    self.selections[sel_key]["selected_links"] = self.selections[
                        sel_key
                    ]["links"]

                self.selections[sel_key]["selection_found"] = True
                # Return pandas.Series of links_ids
                return self.selections[sel_key]["selected_links"].index.tolist()
            else:
                WranglerLogger.error(
                    "Couldn't find path from {} to {}".format(A_id, B_id)
                )
                raise ValueError
        else:
            # unique identifier exists and no need to go through big search
            self.selections[sel_key]["selected_links"] = self.selections[sel_key][
                "candidate_links"
            ]
            self.selections[sel_key]["selection_found"] = True

            return self.selections[sel_key]["selected_links"].index.tolist()

    def validate_properties(
        self,
        properties: dict,
        ignore_existing: bool = False,
        require_existing_for_change: bool = False,
    ) -> bool:
        """
        If there are change or existing commands, make sure that that
        property exists in the network.

        Args:
            properties : properties dictionary to be evaluated
            ignore_existing: If True, will only warn about properties
                that specify an "existing" value.  If False, will fail.
            require_existing_for_change: If True, will fail if there isn't
                a specified value in theproject card for existing when a
                change is specified.

        Returns: boolean value as to whether the properties dictonary is valid.
        """

        validation_error_message = []

        for p in properties:
            if p["property"] not in self.links_df.columns:
                if p.get("change"):
                    validation_error_message.append(
                        '"Change" is specified for attribute {}, but doesn\'t exist in base network\n'.format(
                            p["property"]
                        )
                    )

                if p.get("existing") and not ignore_existing:
                    validation_error_message.append(
                        '"Existing" is specified for attribute {}, but doesn\'t exist in base network\n'.format(
                            p["property"]
                        )
                    )
                elif p.get("existing"):
                    WranglerLogger.warning(
                        '"Existing" is specified for attribute {}, but doesn\'t exist in base network\n'.format(
                            p["property"]
                        )
                    )

            if p.get("change") and not p.get("existing"):
                if require_existing_for_change:
                    validation_error_message.append(
                        '"Change" is specified for attribute {}, but there isn\'t a value for existing.\nTo proceed, run with the setting require_existing_for_change=False'.format(
                            p["property"]
                        )
                    )
                else:
                    WranglerLogger.warning(
                        '"Change" is specified for attribute {}, but there isn\'t a value for existing.\n'.format(
                            p["property"]
                        )
                    )

        if validation_error_message:
            WranglerLogger.error(" ".join(validation_error_message))
            raise ValueError()

    def apply(self, project_card_dictionary: dict):
        """
        Wrapper method to apply a project to a roadway network.

        Args:
            project_card_dictionary: dict
              a dictionary of the project card object

        """

        WranglerLogger.info(
            "Applying Project to Roadway Network: {}".format(
                project_card_dictionary["project"]
            )
        )

        def _apply_individual_change(project_dictionary: dict):

            if project_dictionary["category"].lower() == "roadway property change":
                self.apply_roadway_feature_change(
                    self.select_roadway_features(project_dictionary["facility"]),
                    project_dictionary["properties"],
                )
            elif project_dictionary["category"].lower() == "parallel managed lanes":
                self.apply_managed_lane_feature_change(
                    self.select_roadway_features(project_dictionary["facility"]),
                    project_dictionary["properties"],
                )
            elif project_dictionary["category"].lower() == "add new roadway":
                self.add_new_roadway_feature_change(
                    project_dictionary.get("links"), project_dictionary.get("nodes")
                )
            elif project_dictionary["category"].lower() == "roadway deletion":
                self.delete_roadway_feature_change(
                    project_dictionary.get("links"), project_dictionary.get("nodes")
                )
            else:
                raise (BaseException)

        if project_card_dictionary.get("changes"):
            for project_dictionary in project_card_dictionary["changes"]:
                _apply_individual_change(project_dictionary)
        else:
            _apply_individual_change(project_card_dictionary)

    def apply_roadway_feature_change(
        self, link_idx: list, properties: dict, in_place: bool = True
    ) -> Union(None, RoadwayNetwork):
        """
        Changes the roadway attributes for the selected features based on the
        project card information passed

        Args:
            link_idx : list
                lndices of all links to apply change to
            properties : list of dictionarys
                roadway properties to change
            in_place: boolean
                update self or return a new roadway network object
        """

        # check if there are change or existing commands that that property
        #   exists in the network
        # if there is a set command, add that property to network
        self.validate_properties(properties)

        for i, p in enumerate(properties):
            attribute = p["property"]

            # if project card specifies an existing value in the network
            #   check and see if the existing value in the network matches
            if p.get("existing"):
                network_values = self.links_df.loc[link_idx, attribute].tolist()
                if not set(network_values).issubset([p.get("existing")]):
                    WranglerLogger.warning(
                        "Existing value defined for {} in project card does "
                        "not match the value in the roadway network for the "
                        "selected links".format(attribute)
                    )

            if in_place:
                if "set" in p.keys():
                    self.links_df.loc[link_idx, attribute] = p["set"]
                else:
                    self.links_df.loc[link_idx, attribute] = (
                        self.links_df.loc[link_idx, attribute] + p["change"]
                    )
            else:
                if i == 0:
                    updated_network = copy.deepcopy(self)

                if "set" in p.keys():
                    updated_network.links_df.loc[link_idx, attribute] = p["set"]
                else:
                    updated_network.links_df.loc[link_idx, attribute] = (
                        updated_network.links_df.loc[link_idx, attribute] + p["change"]
                    )

                if i == len(properties) - 1:
                    return updated_network

    def apply_managed_lane_feature_change(
        self, link_idx: list, properties: dict, in_place: bool = True
    ) -> Union(None, RoadwayNetwork):
        """
        Apply the managed lane feature changes to the roadway network

        Args:
            link_idx : list of lndices of all links to apply change to
            properties : list of dictionarys roadway properties to change
            in_place: boolean to indicate whether to update self or return
                a new roadway network object
        """

        # flag ML links
        self.links_df["ML"] = 0
        self.links_df.loc[link_idx, "ML"] = 1

        for p in properties:
            attribute = p["property"]

            if "group" in p.keys():
                attr_value = {}
                attr_value["default"] = p["set"]
                attr_value["timeofday"] = []
                for g in p["group"]:
                    category = g["category"]
                    for tod in g["timeofday"]:
                        attr_value["timeofday"].append(
                            {
                                "category": category,
                                "time": parse_time_spans(tod["time"]),
                                "value": tod["set"],
                            }
                        )

            elif "timeofday" in p.keys():
                attr_value = {}
                attr_value["default"] = p["set"]
                attr_value["timeofday"] = []
                for tod in p["timeofday"]:
                    attr_value["timeofday"].append(
                        {"time": parse_time_spans(tod["time"]), "value": tod["set"]}
                    )

            elif "set" in p.keys():
                attr_value = p["set"]

            else:
                attr_value = ""

            # TODO: decide on connectors info when they are more specific in project card
            if attribute == "ML_ACCESS" and attr_value == "all":
                attr_value = 1

            if attribute == "ML_EGRESS" and attr_value == "all":
                attr_value = 1

            if in_place:
                self.links_df.loc[link_idx, attribute] = attr_value
            else:
                if i == 0:
                    updated_network = copy.deepcopy(self)

                updated_network.links_df.loc[link_idx, attribute] = attr_value

                if i == len(properties) - 1:
                    return updated_network

    def add_new_roadway_feature_change(self, links: dict, nodes: dict) -> None:
        """
        add the new roadway features defined in the project card

        args:
        links : list of dictionaries
        nodes : list of dictionaries
        """

        # TODO:
        # validate links dictonary

        # CHECKS:
        # check if new link model_link_id already exists?
        # check if u and v nodes are already present or not?

        def _add_dict_to_df(df, new_dict):
            df_column_names = df.columns
            new_row_to_add = {}

            # add the fields from project card that are in the network
            for property in df_column_names:
                if property in new_dict.keys():
                    if df[property].dtype == np.float64:
                        value = pd.to_numeric(new_dict[property], downcast="float")
                    elif df[property].dtype == np.int64:
                        value = pd.to_numeric(new_dict[property], downcast="integer")
                    else:
                        value = str(new_dict[property])
                else:
                    value = ""

                new_row_to_add[property] = value

            # add the fields from project card that are NOT in the network
            for key, value in new_dict.items():
                if key not in df_column_names:
                    new_row_to_add[key] = new_dict[key]

            out_df = df.append(new_row_to_add, ignore_index=True)
            return out_df

        if nodes is not None:
            for node in nodes:
                self.nodes_df = _add_dict_to_df(self.nodes_df, node)

        if links is not None:
            for link in links:
                for key in RoadwayNetwork.LINK_FOREIGN_KEY:
                    if link.get(key) is None:
                        msg = "New link to add doesn't contain link foreign key identifier: {}".format(
                            key
                        )
                        WranglerLogger.error(msg)
                        raise ValueError(msg)

                ab_query = "A == " + str(link["A"]) + " and B == " + str(link["B"])

                if not self.links_df.query(ab_query, engine="python").empty:
                    msg = "Link with A = {} and B = {} already exist in the network".format(
                        link["A"], link["B"]
                    )
                    WranglerLogger.error(msg)
                    raise ValueError(msg)

            for link in links:
                self.links_df = _add_dict_to_df(self.links_df, link)

    def delete_roadway_feature_change(
        self, links: dict, nodes: dict, ignore_missing=True
    ) -> None:
        """
        delete the roadway features defined in the project card

        Args:
            links : dict
                list of dictionaries
            nodes : dict
                list of dictionaries
            ignore_missing: bool
                If True, will only warn about links/nodes that are missing from
                network but specified to "delete" in project card
                If False, will fail.
        """

        missing_error_message = []

        if links is not None:
            for key, val in links.items():
                missing_links = [v for v in val if v not in self.links_df[key].tolist()]
                if missing_links:
                    message = "Links attribute {} with values as {} does not exist in the network\n".format(
                        key, missing_links
                    )
                    if ignore_missing:
                        WranglerLogger.warning(message)
                    else:
                        missing_error_message.append(message)

                self.links_df = self.links_df[~self.links_df[key].isin(val)]

        if nodes is not None:
            for key, val in nodes.items():
                missing_nodes = [v for v in val if v not in self.nodes_df[key].tolist()]
                if missing_nodes:
                    message = "Nodes attribute {} with values as {} does not exist in the network\n".format(
                        key, missing_links
                    )
                    if ignore_missing:
                        WranglerLogger.warning(message)
                    else:
                        missing_error_message.append(message)

                self.nodes_df = self.nodes_df[~self.nodes_df[key].isin(val)]

        if missing_error_message:
            WranglerLogger.error(" ".join(missing_error_message))
            raise ValueError()

    def get_property_by_time_period_and_group(
        self, property, time_period=None, category=None
    ):
        """
        Return a series for the properties with a specific group or time period.

        args
        ------
        property: str
          the variable that you want from network
        time_period: list(str)
          the time period that you are querying for
          i.e. ['16:00', '19:00']
        category: str or list(str)(Optional)
          the group category
          i.e. "sov"

          or

          list of group categories in order of search, i.e.
          ["hov3","hov2"]

        returns
        --------
        pandas series
        """

        def _get_property(
            v,
            time_spans=None,
            category=None,
            return_partial_match: bool = False,
            partial_match_minutes: int = 60,
        ):

            if category and not time_spans:
                WranglerLogger.error(
                    "\nShouldn't have a category group without time spans"
                )
                raise ValueError("Shouldn't have a category group without time spans")

            # simple case
            if type(v) in (int, float, str):
                return v

            if not category:
                category = ["default"]
            elif isinstance(category, str):
                category = [category]
            search_cats = [c.lower() for c in category]

            # if no time or group specified, but it is a complex link situation
            if not time_spans:
                if "default" in v.keys():
                    return v["default"]
                else:
                    WranglerLogger.debug("variable: ".format(v))
                    msg = "Variable {} is more complex in network than query".format(v)
                    WranglerLogger.error(msg)
                    raise ValueError(msg)

            if v.get("timeofday"):
                categories = []
                for tg in v["timeofday"]:
                    if (time_spans[0] >= tg["time"][0]) and (
                        time_spans[1] <= tg["time"][1]
                    ):
                        if tg.get("category"):
                            categories += tg["category"]
                            for c in search_cats:
                                print("CAT:", c, tg["category"])
                                if c in tg["category"]:
                                    # print("Var:", v)
                                    # print(
                                    #    "RETURNING:", time_spans, category, tg["value"]
                                    # )
                                    return tg["value"]
                        else:
                            # print("Var:", v)
                            # print("RETURNING:", time_spans, category, tg["value"])
                            return tg["value"]

                    # if there isn't a fully matched time period, see if there is an overlapping one
                    # right now just return the first overlapping ones
                    # TODO return the time period with the largest overlap

                    if (
                        (time_spans[0] >= tg["time"][0])
                        and (time_spans[0] <= tg["time"][1])
                    ) or (
                        (time_spans[1] >= tg["time"][0])
                        and (time_spans[1] <= tg["time"][1])
                    ):
                        overlap_minutes = max(
                            0,
                            min(tg["time"][1], time_spans[1])
                            - max(time_spans[0], tg["time"][0]),
                        )
                        # print("OLM",overlap_minutes)
                        if not return_partial_match and overlap_minutes > 0:
                            WranglerLogger.debug(
                                "Couldn't find time period consistent with {}, but found a partial match: {}. Consider allowing partial matches using 'return_partial_match' keyword or updating query.".format(
                                    time_spans, tg["time"]
                                )
                            )
                        elif (
                            overlap_minutes < partial_match_minutes
                            and overlap_minutes > 0
                        ):
                            WranglerLogger.debug(
                                "Time period: {} overlapped less than the minimum number of minutes ({}<{}) to be considered a match with time period in network: {}.".format(
                                    time_spans,
                                    overlap_minutes,
                                    partial_match_minutes,
                                    tg["time"],
                                )
                            )
                        elif overlap_minutes > 0:
                            WranglerLogger.debug(
                                "Returning a partial time period match. Time period: {} overlapped the minimum number of minutes ({}>={}) to be considered a match with time period in network: {}.".format(
                                    time_spans,
                                    overlap_minutes,
                                    partial_match_minutes,
                                    tg["time"],
                                )
                            )
                            if tg.get("category"):
                                categories += tg["category"]
                                for c in search_cats:
                                    print("CAT:", c, tg["category"])
                                    if c in tg["category"]:
                                        # print("Var:", v)
                                        # print(
                                        #    "RETURNING:",
                                        #    time_spans,
                                        #    category,
                                        #    tg["value"],
                                        # )
                                        return tg["value"]
                            else:
                                # print("Var:", v)
                                # print("RETURNING:", time_spans, category, tg["value"])
                                return tg["value"]

                """
                WranglerLogger.debug(
                    "\nCouldn't find time period for {}, returning default".format(
                        str(time_spans)
                    )
                )
                """
                if "default" in v.keys():
                    # print("Var:", v)
                    # print("RETURNING:", time_spans, v["default"])
                    return v["default"]
                else:
                    # print("Var:", v)
                    WranglerLogger.error(
                        "\nCan't find default; must specify a category in {}".format(
                            str(categories)
                        )
                    )
                    raise ValueError(
                        "Can't find default, must specify a category in: {}".format(
                            str(categories)
                        )
                    )

        time_spans = parse_time_spans(time_period)

        return self.links_df[property].apply(
            _get_property, time_spans=time_spans, category=category
        )

    def create_dummy_connector_links(gp_df: GeoDataFrame, ml_df: GeoDataFrame):
        """
        create dummy connector links between the general purpose and managed lanes

        args:
            gp_df : GeoDataFrame
                dataframe of general purpose links (where managed lane also exists)
            ml_df : GeoDataFrame
                dataframe of corresponding managed lane links
        """

        gp_ml_links_df = pd.concat(
            [gp_df, gp_df.add_prefix("ML_")], axis=1, join="inner"
        )

        access_df = pd.DataFrame(columns=gp_df.columns.values.tolist())
        egress_df = pd.DataFrame(columns=gp_df.columns.values.tolist())

        def _get_connector_references(ref_1: list, ref_2: list, type: str):
            if type == "access":
                out_location_reference = [
                    {"sequence": 1, "point": ref_1[0]["point"]},
                    {"sequence": 2, "point": ref_2[0]["point"]},
                ]

            if type == "egress":
                out_location_reference = [
                    {"sequence": 1, "point": ref_2[1]["point"]},
                    {"sequence": 2, "point": ref_1[1]["point"]},
                ]
            return out_location_reference

        for index, row in gp_ml_links_df.iterrows():
            access_row = {}
            access_row["A"] = row["A"]
            access_row["B"] = row["ML_A"]
            access_row["lanes"] = 1
            access_row["model_link_id"] = (
                row["model_link_id"] + row["ML_model_link_id"] + 1
            )
            access_row["access"] = row["ML_access"]
            access_row["drive_access"] = row["drive_access"]
            access_row["locationReferences"] = _get_connector_references(
                row["locationReferences"], row["ML_locationReferences"], "access"
            )
            access_row["distance"] = haversine_distance(
                access_row["locationReferences"][0]["point"],
                access_row["locationReferences"][1]["point"],
            )
            access_row["roadway"] = "ml_access"
            access_row["name"] = row["name"]
            access_row["ref"] = row["ref"]
            access_df = access_df.append(access_row, ignore_index=True)

            egress_row = {}
            egress_row["A"] = row["ML_B"]
            egress_row["B"] = row["B"]
            egress_row["lanes"] = 1
            egress_row["model_link_id"] = (
                row["model_link_id"] + row["ML_model_link_id"] + 2
            )
            egress_row["access"] = row["ML_access"]
            egress_row["drive_access"] = row["drive_access"]
            egress_row["locationReferences"] = _get_connector_references(
                row["locationReferences"], row["ML_locationReferences"], "egress"
            )
            egress_row["distance"] = haversine_distance(
                egress_row["locationReferences"][0]["point"],
                egress_row["locationReferences"][1]["point"],
            )
            egress_row["roadway"] = "ml_egress"
            egress_row["name"] = row["name"]
            egress_row["ref"] = row["ref"]
            egress_df = egress_df.append(egress_row, ignore_index=True)

        return (access_df, egress_df)

    def create_managed_lane_network(self, in_place: bool=False) -> RoadwayNetwork:
        """
        Create a roadway network with managed lanes links separated out

        args:
            in_place: update self or return a new roadway network object

        returns: A RoadwayNetwork instance
        """

        WranglerLogger.info("Creating network with duplicated managed lanes")

        # TODO make this a more rigorous test
        if "ml_access" in self.links_df["roadway"].tolist():
            msg = "managed lane access links already exist in network; shouldn't be running create managed lane network. Returning network as-is."
            WranglerLogger.error(msg)
            if in_place:
                return
            else:
                return copy.deepcopy(self)

        link_attributes = self.links_df.columns.values.tolist()
        ml_attributes = [i for i in link_attributes if i.startswith("ML_")]

        non_ml_links_df = self.links_df[self.links_df["ML"] == 0]
        non_ml_links_df = non_ml_links_df.drop(ml_attributes, axis=1)

        ml_links_df = self.links_df[self.links_df["ML"] == 1]
        gp_links_df = ml_links_df.drop(ml_attributes, axis=1)

        for attr in link_attributes:
            if attr in ml_attributes and attr not in ["ML_ACCESS", "ML_EGRESS"]:
                gp_attr = attr.split("_")[1]
                ml_links_df[gp_attr] = ml_links_df[attr]

            if (
                attr not in RoadwayNetwork.KEEP_SAME_ATTRIBUTES_ML_AND_GP
                and attr not in RoadwayNetwork.MANAGED_LANES_REQUIRED_ATTRIBUTES
            ):
                ml_links_df[attr] = ""

        ml_links_df = ml_links_df.drop(ml_attributes, axis=1)

        def _update_location_reference(location_reference: list):
            out_location_reference = copy.deepcopy(location_reference)
            out_location_reference[0]["point"] = offset_lat_lon(
                out_location_reference[0]["point"]
            )
            out_location_reference[1]["point"] = offset_lat_lon(
                out_location_reference[1]["point"]
            )
            return out_location_reference

        def _get_line_string(location_reference: list):
            return LineString(
                [location_reference[0]["point"], location_reference[1]["point"]]
            )

        ml_links_df["A"] = (
            ml_links_df["A"] + RoadwayNetwork.MANAGED_LANES_NODE_ID_SCALAR
        )
        ml_links_df["B"] = (
            ml_links_df["B"] + RoadwayNetwork.MANAGED_LANES_NODE_ID_SCALAR
        )
        ml_links_df["model_link_id"] = (
            ml_links_df["model_link_id"] + RoadwayNetwork.MANAGED_LANES_LINK_ID_SCALAR
        )
        ml_links_df["locationReferences"] = ml_links_df["locationReferences"].apply(
            lambda x: _update_location_reference(x)
        )
        ml_links_df["geometry"] = ml_links_df["locationReferences"].apply(
            lambda x: _get_line_string(x)
        )

        access_links_df, egress_links_df = RoadwayNetwork.create_dummy_connector_links(
            gp_links_df, ml_links_df
        )
        access_links_df["geometry"] = access_links_df["locationReferences"].apply(
            lambda x: _get_line_string(x)
        )
        egress_links_df["geometry"] = egress_links_df["locationReferences"].apply(
            lambda x: _get_line_string(x)
        )

        new_links_df = gp_links_df.append(ml_links_df)
        new_links_df = new_links_df.append(access_links_df)
        new_links_df = new_links_df.append(egress_links_df)
        new_links_df = new_links_df.append(non_ml_links_df)
        new_links_df = new_links_df.drop("ML", axis=1)

        # new_links_df["geometry"] = new_links_df["locationReferences"].apply(
        #     lambda x: _get_line_string(x)
        # )

        # only the ml_links_df has the new nodes added
        added_a_nodes = ml_links_df["A"]
        added_b_nodes = ml_links_df["B"]

        new_nodes_df = self.nodes_df

        for a_node in added_a_nodes:
            new_nodes_df = new_nodes_df.append(
                {
                    "model_node_id": a_node,
                    "geometry": Point(
                        new_links_df[new_links_df["A"] == a_node].iloc[0][
                            "locationReferences"
                        ][0]["point"]
                    ),
                    "drive_node": 1,
                },
                ignore_index=True,
            )

        for b_node in added_b_nodes:
            if b_node not in new_nodes_df["model_node_id"].tolist():
                new_nodes_df = new_nodes_df.append(
                    {
                        "model_node_id": b_node,
                        "geometry": Point(
                            new_links_df[new_links_df["B"] == b_node].iloc[0][
                                "locationReferences"
                            ][1]["point"]
                        ),
                        "drive_node": 1,
                    },
                    ignore_index=True,
                )

        new_nodes_df["x"] = new_nodes_df["geometry"].apply(lambda g: g.x)
        new_nodes_df["y"] = new_nodes_df["geometry"].apply(lambda g: g.y)

        new_shapes_df = self.shapes_df

        # managed lanes, access and egress connectors are new geometry
        for index, row in ml_links_df.iterrows():
            new_shapes_df = new_shapes_df.append(
                {"geometry": row["geometry"]}, ignore_index=True
            )
        for index, row in access_links_df.iterrows():
            new_shapes_df = new_shapes_df.append(
                {"geometry": row["geometry"]}, ignore_index=True
            )
        for index, row in egress_links_df.iterrows():
            new_shapes_df = new_shapes_df.append(
                {"geometry": row["geometry"]}, ignore_index=True
            )

        if in_place:
            self.links_df = new_links_df
            self.nodes_df = new_nodes_df
            self.shapes_df = new_shapes_df
        else:
            out_network = copy.deepcopy(self)
            out_network.links_df = new_links_df
            out_network.nodes_df = new_nodes_df
            return out_network

    @staticmethod
    def get_modal_links_nodes(
        links_df: DataFrame, nodes_df: DataFrame, mode: str = None
    ) -> tuple(DataFrame, DataFrame):
        """Returns nodes and link dataframes for specific mode.

        Args:
            links_df: DataFrame of standard network links
            nodes_df: DataFrame of standard network nodes
            mode: mode of the network, one of `drive`,`transit`,
                `walk`, `bike`

        Returns: tuple of DataFrames for links, nodes filtered by mode
        """

        if mode not in RoadwayNetwork.MODES_TO_NETWORK_LINK_VARIABLES.keys():
            msg = "mode value should be one of {}.".format(
                list(RoadwayNetwork.MODES_TO_NETWORK_LINK_VARIABLES.keys())
            )
            WranglerLogger.error(msg)
            raise ValueError(msg)

        mode_link_variable = RoadwayNetwork.MODES_TO_NETWORK_LINK_VARIABLES[mode]
        mode_node_variable = RoadwayNetwork.MODES_TO_NETWORK_NODE_VARIABLES[mode]

        if mode_link_variable not in links_df.columns:
            msg = "{} not in provided links_df list of columns. Available columns are: \n {}".format(mode_link_variable,links_df.columns)
            WranglerLogger.error(msg)

        if mode_node_variable not in nodes_df.columns:
            msg = "{} not in provided nodes_df list of columns. Available columns are: \n {}".format(mode_node_variable,nodes_df.columns)
            WranglerLogger.error(msg)

        modal_links_df = links_df[links_df[mode_link_variable] == 1]
        ##TODO right now we don't filter the nodes because transit-only
        # links with walk access are not marked as having walk access
        # Issue discussed in https://github.com/wsp-sag/network_wrangler/issues/145
        #modal_nodes_df = nodes_df[nodes_df[mode_node_variable] == 1]
        modal_nodes_df = nodes_df

        return modal_links_df, modal_nodes_df

    @staticmethod
    def get_modal_graph(links_df: DataFrame, nodes_df: DataFrame, mode: str = None):
        """Determines if the network graph is "strongly" connected
        A graph is strongly connected if each vertex is reachable from every other vertex.

        Args:
            links_df: DataFrame of standard network links
            nodes_df: DataFrame of standard network nodes
            mode: mode of the network, one of `drive`,`transit`,
                `walk`, `bike`

        Returns: networkx: osmnx: DiGraph  of network
        """
        if mode not in RoadwayNetwork.MODES_TO_NETWORK_LINK_VARIABLES.keys():
            msg = "mode value should be one of {}.".format(
                list(RoadwayNetwork.MODES_TO_NETWORK_LINK_VARIABLES.keys())
            )
            WranglerLogger.error(msg)
            raise ValueError(msg)

        _links_df, _nodes_df = RoadwayNetwork.get_modal_links_nodes(
            links_df, nodes_df, mode=mode
        )
        G = RoadwayNetwork.ox_graph(_nodes_df, _links_df)

        return G

    def is_network_connected(
        self, mode: str = None, links_df: DataFrame = None, nodes_df: DataFrame = None
    ):
        """
        Determines if the network graph is "strongly" connected
        A graph is strongly connected if each vertex is reachable from every other vertex.

        Args:
            mode:  mode of the network, one of `drive`,`transit`,
                `walk`, `bike`
            links_df: DataFrame of standard network links
            nodes_df: DataFrame of standard network nodes

        Returns: boolean
        """

        _nodes_df = nodes_df if nodes_df else self.nodes_df
        _links_df = links_df if links_df else self.links_df

        if mode:
            _links_df, _nodes_df = RoadwayNetwork.get_modal_links_nodes(
                _links_df, _nodes_df, mode=mode
            )
        else:
            WranglerLogger.info(
                "Assessing connectivity without a mode\
                specified. This may have limited value in interpretation.\
                To add mode specificity, add the keyword `mode =` to calling\
                this method"
            )

        # TODO: consider caching graphs if they start to take forever
        #      and we are calling them more than once.
        G = RoadwayNetwork.ox_graph(_nodes_df, _links_df)
        is_connected = nx.is_strongly_connected(G)

        return is_connected

    def assess_connectivity(
        self,
        mode: str = None,
        ignore_end_nodes: bool = True,
        links_df: DataFrame = None,
        nodes_df: DataFrame = None,
    ):
        """Returns a network graph and list of disconnected subgraphs
        as described by a list of their member nodes.

        Args:
            mode:  mode of the network, one of `drive`,`transit`,
                `walk`, `bike`
            ignore_end_nodes: if True, ignores stray singleton nodes
            links_df: if specified, will assess connectivity of this
                links list rather than self.links_df
            nodes_df: if specified, will assess connectivity of this
                nodes list rather than self.nodes_df

        Returns: Tuple of
            Network Graph (osmnx flavored networkX DiGraph)
            List of disconnected subgraphs described by the list of their
                member nodes (as described by their `model_node_id`)
        """
        _nodes_df = nodes_df if nodes_df else self.nodes_df
        _links_df = links_df if links_df else self.links_df

        if mode:
            _links_df, _nodes_df = RoadwayNetwork.get_modal_links_nodes(
                _links_df, _nodes_df, mode=mode
            )
        else:
            WranglerLogger.info(
                "Assessing connectivity without a mode\
                specified. This may have limited value in interpretation.\
                To add mode specificity, add the keyword `mode =` to calling\
                this method"
            )

        G = RoadwayNetwork.ox_graph(_nodes_df, _links_df)
        # sub_graphs = [s for s in sorted(nx.strongly_connected_component_subgraphs(G), key=len, reverse=True)]
        sub_graphs = [
            s
            for s in sorted(
                (G.subgraph(c) for c in nx.strongly_connected_components(G)),
                key=len,
                reverse=True,
            )
        ]

        sub_graph_nodes = [
            list(s)
            for s in sorted(nx.strongly_connected_components(G), key=len, reverse=True)
        ]

        # sorted on decreasing length, dropping the main sub-graph
        disconnected_sub_graph_nodes = sub_graph_nodes[1:]

        # dropping the sub-graphs with only 1 node
        if ignore_end_nodes:
            disconnected_sub_graph_nodes = [
                list(s) for s in disconnected_sub_graph_nodes if len(s) > 1
            ]

        WranglerLogger.info(
            "{} for disconnected networks for mode = {}:\n{}".format(
                RoadwayNetwork.NODE_FOREIGN_KEY,
                mode,
                "\n".join(list(map(str,disconnected_sub_graph_nodes))),
            )
        )
        return G, disconnected_sub_graph_nodes

    @staticmethod
    def network_connection_plot(G, disconnected_subgraph_nodes: list):
        """Plot a graph to check for network connection.

        Args:
            G: OSMNX flavored networkX graph.
            disconnected_subgraph_nodes: List of disconnected subgraphs described by the list of their
                member nodes (as described by their `model_node_id`).

        returns: fig, ax : tuple
        """

        colors = []
        for i in range(len(disconnected_subgraph_nodes)):
            colors.append("#%06X" % randint(0, 0xFFFFFF))

        fig, ax = ox.plot_graph(
            G,
            fig_height=16,
            fig_width=16,
            show=False,
            close=True,
            edge_color="black",
            edge_alpha=0.1,
            node_color="black",
            node_alpha=0.5,
            node_size=10,
        )
        i = 0
        for nodes in disconnected_subgraph_nodes:
            for n in nodes:
                size = 100
                ax.scatter(G.nodes[n]["x"], G.nodes[n]["y"], c=colors[i], s=size)
            i = i + 1

        return fig, ax

    def selection_map(selection: tuple):
        """
        Shows which links are selected for roadway property change or parallel
        managed lanes category of roadway projects
        """

        sel_key = selection[0]
        sel_val = selection[1]

        A_node = sel_key[1]
        B_node = sel_key[2]
        selected_graph = sel_val["graph"]
        selected_links = sel_val["selected_links"]

        m = ox.plot_graph_folium(
            selected_graph, edge_color=None, tiles="cartodbpositron"
        )

        def _folium_node(node, color="white", icon=""):
            node_marker = folium.Marker(
                location=[node["y"], node["x"]],
                icon=folium.Icon(icon=icon, color=color),
            )
            return node_marker

        A = selected_graph.nodes[A_node]
        B = selected_graph.nodes[B_node]

        _folium_node(A, color="green", icon="play").add_to(m)
        _folium_node(B, color="red", icon="star").add_to(m)

        for _, row in selected_links.iterrows():
            pl = ox.plot.make_folium_polyline(
                edge=row, edge_color="blue", edge_width=5, edge_opacity=0.8
            )
            pl.add_to(m)

        return m

    def deletion_map(self, links: dict, nodes: dict):
        """
        Shows which links and nodes are deleted from the roadway network
        """
        # deleted_links = None
        # deleted_nodes = None

        missing_error_message = []

        if links is not None:
            for key, val in links.items():
                deleted_links = self.links_df[self.links_df[key].isin(val)]

                node_list_foreign_keys = list(
                    set(
                        [
                            i
                            for fk in RoadwayNetwork.LINK_FOREIGN_KEY
                            for i in list(deleted_links[fk])
                        ]
                    )
                )
                candidate_nodes = self.nodes_df.loc[node_list_foreign_keys]
        else:
            deleted_links = None

        if nodes is not None:
            for key, val in nodes.items():
                deleted_nodes = self.nodes_df[self.nodes_df[key].isin(val)]
        else:
            deleted_nodes = None

        G = RoadwayNetwork.ox_graph(candidate_nodes, deleted_links)

        m = ox.plot_graph_folium(G, edge_color="red", tiles="cartodbpositron")

        def _folium_node(node, color="white", icon=""):
            node_circle = folium.Circle(
                location=[node["y"], node["x"]],
                radius=2,
                fill=True,
                color=color,
                fill_opacity=0.8,
            )
            return node_circle

        if deleted_nodes is not None:
            for _, row in deleted_nodes.iterrows():
                _folium_node(row, color="red").add_to(m)

        return m

    def addition_map(self, links: dict, nodes: dict):
        """
        Shows which links and nodes are added to the roadway network
        """

        if links is not None:
            link_ids = []
            for link in links:
                link_ids.append(link.get(RoadwayNetwork.UNIQUE_LINK_KEY))

            added_links = self.links_df[
                self.links_df[RoadwayNetwork.UNIQUE_LINK_KEY].isin(link_ids)
            ]
            node_list_foreign_keys = list(
                set(
                    [
                        i
                        for fk in RoadwayNetwork.LINK_FOREIGN_KEY
                        for i in list(added_links[fk])
                    ]
                )
            )
            try:
                candidate_nodes = self.nodes_df.loc[node_list_foreign_keys]
            except:
                return None

        if nodes is not None:
            node_ids = []
            for node in nodes:
                node_ids.append(node.get(RoadwayNetwork.UNIQUE_NODE_KEY))

            added_nodes = self.nodes_df[
                self.nodes_df[RoadwayNetwork.UNIQUE_NODE_KEY].isin(node_ids)
            ]
        else:
            added_nodes = None

        G = RoadwayNetwork.ox_graph(candidate_nodes, added_links)

        m = ox.plot_graph_folium(G, edge_color="green", tiles="cartodbpositron")

        def _folium_node(node, color="white", icon=""):
            node_circle = folium.Circle(
                location=[node["y"], node["x"]],
                radius=2,
                fill=True,
                color=color,
                fill_opacity=0.8,
            )
            return node_circle

        if added_nodes is not None:
            for _, row in added_nodes.iterrows():
                _folium_node(row, color="green").add_to(m)

        return m



import pandas as pd
from .Logger import WranglerLogger
import math
import copy


def point_df_to_geojson(df: pd.DataFrame, properties: list):
    """
    Author: Geoff Boeing:
    https://geoffboeing.com/2015/10/exporting-python-data-geojson/
    """
    from .RoadwayNetwork import RoadwayNetwork

    geojson = {"type": "FeatureCollection", "features": []}
    for _, row in df.iterrows():
        feature = {
            "type": "Feature",
            "properties": {},
            "geometry": {"type": "Point", "coordinates": []},
        }
        feature["geometry"]["coordinates"] = [row["geometry"].x, row["geometry"].y]
        feature["properties"][RoadwayNetwork.NODE_FOREIGN_KEY] = row.name
        for prop in properties:
            feature["properties"][prop] = row[prop]
        geojson["features"].append(feature)
    return geojson


def link_df_to_json(df: pd.DataFrame, properties: list):
    """
    Modified from: Geoff Boeing:
    https://geoffboeing.com/2015/10/exporting-python-data-geojson/
    """
    df["distance"].fillna(0)
    json = []
    for _, row in df.iterrows():
        feature = {}
        for prop in properties:
            feature[prop] = row[prop]
        json.append(feature)
    return json


def topological_sort(adjacency_list, visited_list):
    """
    Topological sorting for Acyclic Directed Graph
    """

    output_stack = []

    def topology_sort_util(vertex):
        if not visited_list[vertex]:
            visited_list[vertex] = True
            for neighbor in adjacency_list[vertex]:
                topology_sort_util(neighbor)
            output_stack.insert(0, vertex)

    for vertex in visited_list:
        topology_sort_util(vertex)

    return output_stack


def make_slug(text, delimiter: str = "_"):
    """
    makes a slug from text
    """
    import re

    text = re.sub("[,.;@#?!&$']+", "", text.lower())
    return re.sub("[\ ]+", delimiter, text)


def parse_time_spans(times):
    """
    parse time spans into tuples of seconds from midnight
    can also be used as an apply function for a pandas series
    Parameters
    -----------
    times: tuple(string) or tuple(int) or list(string) or list(int)

    returns
    --------
    tuple(integer)
      time span as seconds from midnight
    """
    try:
        start_time, end_time = times
    except:
        msg = "ERROR: times should be a tuple or list of two, got: {}".format(times)
        WranglerLogger.error(msg)
        raise ValueError(msg)

    # If times are strings, convert to int in seconds, else return as ints
    if isinstance(start_time, str) and isinstance(end_time, str):
        start_time = start_time.strip()
        end_time = end_time.strip()

        # If time is given without seconds, add 00
        if len(start_time) <= 5:
            start_time += ":00"
        if len(end_time) <= 5:
            end_time += ":00"

        # Convert times to seconds from midnight (Partride's time storage)
        h0, m0, s0 = start_time.split(":")
        start_time_sec = int(h0) * 3600 + int(m0) * 60 + int(s0)

        h1, m1, s1 = end_time.split(":")
        end_time_sec = int(h1) * 3600 + int(m1) * 60 + int(s1)

        return (start_time_sec, end_time_sec)

    elif isinstance(start_time, int) and isinstance(end_time, int):
        return times

    else:
        WranglerLogger.error("ERROR: times should be ints or strings")
        raise ValueError()

    return (start_time_sec, end_time_sec)


def offset_lat_lon(lon_lat_point: list, offset_meters=100):
    in_lon = lon_lat_point[0]
    in_lat = lon_lat_point[1]

    # Earth's radius
    radius = 6378137

    # offset in radians
    offset_lat_radians = offset_meters / radius
    offset_lon_radians = offset_meters / (radius * math.cos(math.pi * in_lat / 180))

    # offset lat lon
    latO = in_lat + offset_lat_radians * 180 / math.pi
    lonO = in_lon + offset_lon_radians * 180 / math.pi

    return [lonO, latO]


def haversine_distance(origin: list, destination: list):
    lon1, lat1 = origin
    lon2, lat2 = destination
    radius = 6378137  # meter

    dlat = math.radians(lat2 - lat1)
    dlon = math.radians(lon2 - lon1)
    a = math.sin(dlat / 2) * math.sin(dlat / 2) + math.cos(
        math.radians(lat1)
    ) * math.cos(math.radians(lat2)) * math.sin(dlon / 2) * math.sin(dlon / 2)
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    d = radius * c

    return d



import os
import pytest

slug_test_list = [
    {"text": "I am a roadway", "delim": "_", "answer": "i_am_a_roadway"},
    {"text": "I'm a roadway", "delim": "_", "answer": "im_a_roadway"},
    {"text": "I am a roadway", "delim": "-", "answer": "i-am-a-roadway"},
    {"text": "I am a roadway", "delim": "", "answer": "iamaroadway"},
]


@pytest.mark.parametrize("slug_test", slug_test_list)
def test_get_slug(request, slug_test):
    print("\n--Starting:", request.node.name)
    from network_wrangler.Utils import make_slug

    slug = make_slug(slug_test["text"], delimiter=slug_test["delim"])
    print("From: {} \nTo: {}".format(slug_test["text"], slug))
    print("Expected: {}".format(slug_test["answer"]))
    assert slug == slug_test["answer"]


@pytest.mark.travis
def test_time_convert(request):
    print("\n--Starting:", request.node.name)

    time_tests = [
        (("00:00:00", "00:00:10"), (0, 10)),
        (("0:00", "0:10:00"), (0, 600)),
        (("01:02:03", "01:02:23"), (3723, 3743)),
        (("1:02", "1:02:13"), (3720, 3733)),
        (("25:24:23", "25:24:33"), (91463, 91473)),
        (("250:24:23", "250:24:25"), (901463, 901465)),
    ]

    from pandas import DataFrame

    df = DataFrame(time_tests, columns=["time", "time_results"])
    print("Original Time Series", df)

    from network_wrangler.Utils import parse_time_spans

    df["time"] = df["time"].apply(parse_time_spans)
    print("Result Time Series", df)
    from pandas.testing import assert_series_equal

    assert_series_equal(df["time"], df["time_results"], check_names=False)


null_val_type_list = [("1", 0), ("3.2", 0.0), ("Howdy", ""), ("False", False)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Utility functions for logging

import sys
import logging

__all__ = ["WranglerLogger", "setupLogging"]

WranglerLogger = logging.getLogger("WranglerLogger")


def setupLogging(level=None, logFileName=None, logToConsole=False):
    """
        Sets up the Wrangler Logger
        
    args:
        level (int): the level of logging that will be recorded 
        logFileName (string): the name of the log file that will get created
        logToConsole(boolean): if True, logging will go to the console
        
        """

    if level is None:
        WranglerLogger.setLevel(logging.DEBUG)
    else:
        WranglerLogger.setLevel(level)

    FORMAT = logging.Formatter(
        "%(asctime)-15s %(levelname)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S,"
    )

    if logFileName:
        file_handler = logging.FileHandler(logFileName)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(FORMAT)
        WranglerLogger.addHandler(file_handler)

    if logToConsole:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setLevel(logging.WARNING)
        console_handler.setFormatter(FORMAT)
        WranglerLogger.addHandler(console_handler)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import yaml
import json

from jsonschema import validate
from jsonschema.exceptions import ValidationError
from jsonschema.exceptions import SchemaError
from .Logger import WranglerLogger


class ProjectCard(object):
    """
    Representation of a Project Card

    Attributes:
        __dict__: Dictionary of project card attributes
        valid: Boolean indicating if data conforms to project card data schema
    """

    TRANSIT_CATEGORIES = ["Transit Service Property Change"]

    # categories that may affect transit, but only as a secondary
    # effect of changing roadways
    SECONDARY_TRANSIT_CATEGORIES = ["Roadway Deletion", "Parallel Managed Lanes"]

    ROADWAY_CATEGORIES = [
        "Roadway Property Change",
        "Roadway Deletion",
        "Parallel Managed lanes",
        "Add New Roadway",
    ]

    UNSPECIFIED_PROJECT_NAMES = ["", "TO DO User Define", "USER TO define"]

    def __init__(self, attribute_dictonary: dict):
        """
        Constructor

        args:
        attribute_dictonary: a nested dictionary of attributes
        """
        self.__dict__.update(attribute_dictonary)
        self.valid = False

        # todo more unstructuring of project card yaml

    def __str__(self):
        s = ["{}: {}".format(key, value) for key, value in self.__dict__.items()]
        return "\n".join(s)

    @staticmethod
    def read(path_to_card: str, validate: bool = True):
        """
        Reads and validates a Project card

        args:
        path_to_card: the path to the project card

        Returns a Project Card object
        """

        with open(path_to_card, "r") as cardfile:
            attribute_dictionary = yaml.safe_load(cardfile)
            attribute_dictionary["file"] = path_to_card
            card = ProjectCard(attribute_dictionary)

        if card.project in ProjectCard.UNSPECIFIED_PROJECT_NAMES:
            msg = "Card must have valid project name: {}".format(path_to_card)
            WranglerLogger.error(msg)
            raise ValueError(msg)

        card.valid = False
        if validate:
            card.valid = ProjectCard.validate_project_card_schema(path_to_card)

        return card

    def write(self, filename: str = None):
        """
        Writes project card dictionary to YAML file
        """
        if not filename:
            from network_wrangler.Utils import make_slug

            filename = make_slug(self.project) + ".yml"
        with open(filename, "w") as outfile:
            yaml.dump(self.__dict__, outfile, default_flow_style=False)

    @staticmethod
    def validate_project_card_schema(
        card_file, card_schema_file: str = "project_card.json"
    ) -> bool:
        """
        Tests project card schema validity by evaluating if it conforms to the schemas
        returns: boolean
        """
        if not os.path.exists(card_schema_file):
            base_path = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "schemas"
            )
            card_schema_file = os.path.join(base_path, card_schema_file)

        with open(card_schema_file) as schema_json_file:
            schema = json.load(schema_json_file)

        with open(card_file, "r") as card:
            card_json = yaml.safe_load(card)

        try:
            validate(card_json, schema)
            return True

        except ValidationError as exc:
            WranglerLogger.error("Failed Project Card validation: Validation Error")
            WranglerLogger.error("Project Card File Loc:{}".format(card_file))
            WranglerLogger.error("Project Card Schema Loc:{}".format(card_schema_file))
            WranglerLogger.error(exc.message)

        except SchemaError as exc:
            WranglerLogger.error("Failed Project Card schema validation: Schema Error")
            WranglerLogger.error("Project Card Schema Loc:{}".format(card_schema_file))
            WranglerLogger.error(exc.message)

        except yaml.YAMLError as exc:
            WranglerLogger.error(exc.message)

    @staticmethod
    def build_link_selection_query(
        selection: dict,
        unique_model_link_identifiers: [],
        mode="drive_access",
        ignore=[],
    ):
        sel_query = "("
        count = 0

        selection_keys = [k for l in selection["link"] for k, v in l.items()]
        num_unique_model_link_identifiers = len(
            set(unique_model_link_identifiers).intersection(selection_keys)
        )
        unique_model_link_identifer_exist = num_unique_model_link_identifiers > 0

        for l in selection["link"]:
            for key, value in l.items():

                if key in ignore:
                    continue

                if (
                    unique_model_link_identifer_exist
                    and key not in unique_model_link_identifiers
                ):
                    continue

                count = count + 1

                if isinstance(value, list):
                    sel_query = sel_query + "("
                    v = 1
                    for i in value:  # building an OR query with each element in list
                        if isinstance(i, str):
                            sel_query = sel_query + key + '.str.contains("' + i + '")'
                        else:
                            sel_query = sel_query + key + "==" + str(i)
                        if v != len(value):
                            sel_query = sel_query + " or "
                            v = v + 1
                    sel_query = sel_query + ")"
                else:
                    sel_query = sel_query + key + "==" + '"' + str(value) + '"'

                if not unique_model_link_identifer_exist and count != (
                    len(selection["link"]) - len(ignore)
                ):
                    sel_query = sel_query + " and "

                if (
                    unique_model_link_identifer_exist
                    and count != num_unique_model_link_identifiers
                ):
                    sel_query = sel_query + " and "

        if not unique_model_link_identifer_exist:
            if count > 0:
                sel_query = sel_query + " and "
            sel_query = sel_query + mode + "==1"

        sel_query = sel_query + ")"

        return sel_query

    def roadway_attribute_change(self, card: dict):
        """
        Probably delete.
        Reads a Roadway Attribute Change card.

        args:
        card: the project card stored in a dictionary
        """
        WranglerLogger.info(card.get("Category"))

    def new_roadway(self, card: dict):
        """
        Probably delete.
        Reads a New Roadway card.

        args:
        card: the project card stored in a dictionary
        """
        WranglerLogger.info(card.get("Category"))

    def transit_attribute_change(self, card: dict):
        """
        Probably delete.
        Reads a Transit Service Attribute Change card.

        args:
        card: the project card stored in a dictionary
        """
        WranglerLogger.info(card.get("Category"))

    def new_transit_right_of_way(self, card: dict):
        """
        Probably delete.
        Reads a New Transit Dedicated Right of Way card.

        args:
        card: the project card stored in a dictionary
        """
        WranglerLogger.info(card.get("Category"))

    def parallel_managed_lanes(self, card: dict):
        """
        Probably delete.
        Reads a Parallel Managed lanes card.

        args:
        card: the project card stored in a dictionary
        """
        WranglerLogger.info(card.get("Category"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import copy
import os
import re
from typing import Tuple, Union

import networkx as nx
import numpy as np
import pandas as pd
import partridge as ptg
from partridge.config import default_config

from .Logger import WranglerLogger
from .Utils import parse_time_spans
from .RoadwayNetwork import RoadwayNetwork


class TransitNetwork(object):
    """
    Representation of a Transit Network.

    Usage:
      import network_wrangler as wr
      stpaul = r'/home/jovyan/work/example/stpaul'
      tc=wr.TransitNetwork.read(path=stpaul)

    """

    # PK = primary key, FK = foreign key
    SHAPES_FOREIGN_KEY = "shape_model_node_id"
    STOPS_FOREIGN_KEY = "model_node_id"

    REQUIRED_FILES = [
        "agency.txt",
        "frequencies.txt",
        "routes.txt",
        "shapes.txt",
        "stop_times.txt",
        "stops.txt",
        "trips.txt",
    ]

    def __init__(self, feed: DotDict = None, config: nx.DiGraph = None):
        """
        Constructor
        """
        self.feed: DotDict = feed
        self.config: nx.DiGraph = config
        self.road_net: RoadwayNetwork = None
        self.graph: nx.MultiDiGraph = None
        self.feed_path = None

        self.validated_frequencies = False
        self.validated_road_network_consistency = False

        if not self.validate_frequencies():
            raise ValueError(
                "Transit lines with non-positive frequencies exist in the network"
            )

    @staticmethod
    def read(feed_path: str, fast: bool = False) -> TransitNetwork:
        """
        Read GTFS feed from folder and TransitNetwork object
        """
        config = default_config()
        feed = ptg.load_feed(feed_path, config=config)
        WranglerLogger.info("Read in transit feed from: {}".format(feed_path))
        updated_config = TransitNetwork.validate_feed(feed, config)

        # Read in each feed so we can write over them
        new_feed = DotDict()
        for node in updated_config.nodes.keys():
            # Load (initiate Partridge's lazy load)
            new_feed[node.replace(".txt", "")] = feed.get(node)

        transit_network = TransitNetwork(feed=new_feed, config=updated_config)
        transit_network.feed_path = feed_path
        return transit_network

    @staticmethod
    def validate_feed(feed: DotDict, config: nx.DiGraph) -> bool:
        """
        Since Partridge lazily loads the df, load each file to make sure it
        actually works.

        Partridge uses a DiGraph from the networkx library to represent the
        relationships between GTFS files. Each file is a 'node', and the
        relationship between files are 'edges'.
        """
        updated_config = copy.deepcopy(config)
        files_not_found = []
        for node in config.nodes.keys():

            n = feed.get(node)
            WranglerLogger.debug("...{}:\n{}".format(node, n[:10]))
            if n.shape[0] == 0:
                WranglerLogger.info(
                    "Removing {} from transit network config because file not found".format(
                        node
                    )
                )
                updated_config.remove_node(node)
                if node in TransitNetwork.REQUIRED_FILES:
                    files_not_found.append(node)

        if files_not_found:
            msg = "Required files not found or valid: {}".format(
                ",".join(files_not_found)
            )
            WranglerLogger.error(msg)
            raise AttributeError(msg)
            return False

        return updated_config

    def validate_frequencies(self) -> Bool:
        """
        Validates that there are no transit trips in the feed with zero frequencies.

        Changes state of self.validated_frequencies boolean based on outcome.

        Returns:
            boolean indicating if valid or not.
        """

        _valid = True
        zero_freq = self.feed.frequencies[self.feed.frequencies.headway_secs <= 0]

        if len(zero_freq.index) > 0:
            _valid = False
            msg = "Transit lines {} have non-positive frequencies".format(
                zero_freq.trip_id.to_list()
            )
            WranglerLogger.error(msg)

        self.validated_frequencies = True

        return _valid

    def validate_road_network_consistencies(self) -> Bool:
        """
        Validates transit network against the road network for both stops
        and shapes.

        Returns:
            boolean indicating if valid or not.
        """
        if self.road_net is None:
            raise ValueError(
                "RoadwayNetwork not set yet, see TransitNetwork.set_roadnet()"
            )

        valid = True

        valid_stops = self.validate_transit_stops()
        valid_shapes = self.validate_transit_shapes()

        self.validated_road_network_consistency = True

        if not valid_stops or not valid_shapes:
            valid = False
            raise ValueError("Transit network is not consistent with road network.")

        return valid

    def validate_transit_stops(self) -> Bool:
        """
        Validates that all transit stops are part of the roadway network.

        Returns:
            Boolean indicating if valid or not.
        """

        if self.road_net is None:
            raise ValueError(
                "RoadwayNetwork not set yet, see TransitNetwork.set_roadnet()"
            )

        stops = self.feed.stops
        nodes = self.road_net.nodes_df

        valid = True

        stop_ids = [int(s) for s in stops[TransitNetwork.STOPS_FOREIGN_KEY].to_list()]
        node_ids = [int(n) for n in nodes[RoadwayNetwork.NODE_FOREIGN_KEY].to_list()]

        if not set(stop_ids).issubset(node_ids):
            valid = False
            missing_stops = list(set(stop_ids) - set(node_ids))
            msg = "Not all transit stops are part of the roadyway network. "
            msg += "Missing stops ({}) from the roadway nodes are {}.".format(
                TransitNetwork.STOPS_FOREIGN_KEY, missing_stops
            )
            WranglerLogger.error(msg)

        return valid

    def validate_transit_shapes(self) -> Bool:
        """
        Validates that all transit shapes are part of the roadway network.

        Returns:
            Boolean indicating if valid or not.
        """

        if self.road_net is None:
            raise ValueError(
                "RoadwayNetwork not set yet, see TransitNetwork.set_roadnet()"
            )

        shapes = self.feed.shapes
        nodes = self.road_net.nodes_df

        valid = True

        shape_ids = [
            int(s) for s in shapes[TransitNetwork.SHAPES_FOREIGN_KEY].to_list()
        ]
        node_ids = [int(n) for n in nodes[RoadwayNetwork.NODE_FOREIGN_KEY].to_list()]

        if not set(shape_ids).issubset(node_ids):
            valid = False
            missing_shapes = list(set(shape_ids) - set(node_ids))
            msg = "Not all transit shapes are part of the roadyway network. "
            msg += "Missing shapes ({}) from the roadway network are {}.".format(
                TransitNetwork.SHAPES_FOREIGN_KEY, missing_shapes
            )
            WranglerLogger.error(msg)

        return valid

    @staticmethod
    def route_ids_in_routestxt(feed):
        """
        STUB: Wherever route_id occurs, make sure it is in routes.txt

        Returns:
            Boolean
        """
        pass

    @staticmethod
    def trip_ids_in_tripstxt(feed):
        """
        STUB: Wherever trip_id occurs, make sure it is in trips.txt

        Returns:
            Boolean
        """
        pass

    @staticmethod
    def shape_ids_in_shapestxt(feed):
        """
        STUB: Wherever shape_id occurs, make sure it is in shapes.txt

        Returns:
            Boolean
        """
        pass

    @staticmethod
    def stop_ids_in_stopstxt(feed):
        """
        STUB: Wherever stop_id occurs, make sure it is in stops.txt

        Returns:
            Boolean
        """
        pass

    @staticmethod
    def validate_network_connectivity(feed):
        """
        Validates foreign keys are present in all connecting feed files.


        Returns:
            Boolean
        """
        result = True
        result = result and TransitNetwork.route_ids_in_routestxt(feed)
        result = result and TransitNetwork.trip_ids_in_tripstxt(feed)
        result = result and TransitNetwork.shape_ids_in_shapestxt(feed)
        result = result and TransitNetwork.stop_ids_in_stopstxt(feed)
        return result

    def set_roadnet(
        self,
        road_net: RoadwayNetwork,
        graph_shapes: bool = False,
        graph_stops: bool = False,
        validate_consistency: bool = True,
    ) -> None:
        self.road_net: RoadwayNetwork = road_net
        self.graph: nx.MultiDiGraph = RoadwayNetwork.ox_graph(
            road_net.nodes_df, road_net.links_df
        )
        if graph_shapes:
            self._graph_shapes()
        if graph_stops:
            self._graph_stops()

        if validate_consistency:
            self.validate_road_network_consistencies()

    def _graph_shapes(self) -> None:
        existing_shapes = self.feed.shapes
        # graphed_shapes = pd.DataFrame()

        # for shape_id in shapes:
        # TODO traverse point by point, mapping shortest path on graph,
        # then append to a list
        # return total list of all link ids
        # rebuild rows in shapes dataframe and add to graphed_shapes
        # make graphed_shapes a GeoDataFrame

        # self.feed.shapes = graphed_shapes

    def _graph_stops(self) -> None:
        existing_stops = self.feed.stops
        # graphed_stops = pd.DataFrame()

        # for stop_id in stops:
        # TODO

        # self.feed.stops = graphed_stops

    def write(self, path: str = ".", filename: str = None) -> None:
        """
        Writes a network in the transit network standard

        Parameters
        ------------
        path: the path were the output will be saved
        filename: the name prefix of the transit files that will be generated
        """
        WranglerLogger.info("Writing transit to directory: {}".format(path))
        for node in self.config.nodes.keys():

            df = self.feed.get(node.replace(".txt", ""))
            if not df.empty:
                if filename:
                    outpath = os.path.join(path, filename + "_" + node)
                else:
                    outpath = os.path.join(path, node)
                WranglerLogger.debug("Writing file: {}".format(outpath))

                df.to_csv(outpath, index=False)

    def apply(self, project_card_dictionary: dict):
        """
        Wrapper method to apply a project to a transit network.

        args
        ------
        project_card_dictionary: dict
          a dictionary of the project card object

        """
        WranglerLogger.info(
            "Applying Project to Transit Network: {}".format(
                project_card_dictionary["project"]
            )
        )

        def _apply_individual_change(project_dictionary: dict):
            if (
                project_dictionary["category"].lower()
                == "transit service property change"
            ):
                self.apply_transit_feature_change(
                    self.select_transit_features(project_dictionary["facility"]),
                    project_dictionary["properties"],
                )
            elif project_dictionary["category"].lower() == "parallel managed lanes":
                # Grab the list of nodes in the facility from road_net
                # It should be cached because managed lane projects are
                # processed by RoadwayNetwork first via
                # Scenario.apply_all_projects
                try:
                    managed_lane_nodes = self.road_net.selections(
                        self.road_net.build_selection_key(
                            project_dictionary["facility"]
                        )
                    )["route"]
                except ValueError:
                    WranglerLogger.error(
                        "RoadwayNetwork not set yet, see TransitNetwork.set_roadnet()"
                    )

                # Reroute any transit using these nodes
                self.apply_transit_managed_lane(
                    self.select_transit_features_by_nodes(managed_lane_nodes),
                    managed_lane_nodes,
                )
            elif project_dictionary["category"].lower() == "roadway deletion":
                WranglerLogger.warning(
                    "Roadway Deletion not yet implemented in Transit; ignoring"
                )
            else:
                msg = "{} not implemented yet in TransitNetwork; can't apply.".format(
                    project_dictionary["category"]
                )
                WranglerLogger.error(msg)
                raise (msg)

        if project_card_dictionary.get("changes"):
            for project_dictionary in project_card_dictionary["changes"]:
                _apply_individual_change(project_dictionary)
        else:
            _apply_individual_change(project_card_dictionary)

    def select_transit_features(self, selection: dict) -> pd.Series:
        """
        Selects transit features that satisfy selection criteria

        Args:
            selection : selection dictionary

        Returns: trip identifiers : list of GTFS trip IDs in the selection
        """
        trips = self.feed.trips
        routes = self.feed.routes
        freq = self.feed.frequencies

        # Turn selection's values into lists if they are not already
        for key in selection.keys():
            if type(selection[key]) not in [list, tuple]:
                selection[key] = [selection[key]]

        # Based on the key in selection, filter trips
        if "trip_id" in selection:
            trips = trips[trips.trip_id.isin(selection["trip_id"])]

        elif "route_id" in selection:
            trips = trips[trips.route_id.isin(selection["route_id"])]

        elif "route_short_name" in selection:
            routes = routes[routes.route_short_name.isin(selection["route_short_name"])]
            trips = trips[trips.route_id.isin(routes["route_id"])]

        elif "route_long_name" in selection:
            matches = []
            for sel in selection["route_long_name"]:
                for route_long_name in routes["route_long_name"]:
                    x = re.search(sel, route_long_name)
                    if x is not None:
                        matches.append(route_long_name)

            routes = routes[routes.route_long_name.isin(matches)]
            trips = trips[trips.route_id.isin(routes["route_id"])]

        else:
            WranglerLogger.error("Selection not supported %s", selection.keys())
            raise ValueError

        # If a time key exists, filter trips using frequency table
        if selection.get("time"):
            selection["time"] = parse_time_spans(selection["time"])
        elif selection.get("start_time") and selection.get("end_time"):
            selection["time"] = parse_time_spans(
                [selection["start_time"], selection["end_time"]]
            )
            # Filter freq to trips in selection
            freq = freq[freq.trip_id.isin(trips["trip_id"])]
            freq = freq[freq.start_time == selection["time"][0]]
            freq = freq[freq.end_time == selection["time"][1]]

            # Filter trips table to those still in freq table
            trips = trips[trips.trip_id.isin(freq["trip_id"])]

        # If any other key exists, filter routes or trips accordingly
        for key in selection.keys():
            if key not in [
                "trip_id",
                "route_id",
                "route_short_name",
                "route_long_name",
                "time",
            ]:
                if key in trips:
                    trips = trips[trips[key].isin(selection[key])]
                elif key in routes:
                    routes = routes[routes[key].isin(selection[key])]
                    trips = trips[trips.route_id.isin(routes["route_id"])]
                else:
                    WranglerLogger.error("Selection not supported %s", key)
                    raise ValueError

        # Check that there is at least one trip in trips table or raise error
        if len(trips) < 1:
            WranglerLogger.error("Selection returned zero trips")
            raise ValueError

        # Return pandas.Series of trip_ids
        return trips["trip_id"]

    def select_transit_features_by_nodes(
        self, node_ids: list, require_all: bool = False
    ) -> pd.Series:
        """
        Selects transit features that use any one of a list of node_ids

        Args:
            node_ids: list (generally coming from nx.shortest_path)
            require_all : bool if True, the returned trip_ids must traverse all of
              the nodes (default = False)

        Returns:
            trip identifiers  list of GTFS trip IDs in the selection
        """
        # If require_all, the returned trip_ids must traverse all of the nodes
        # Else, filter any shapes that use any one of the nodes in node_ids
        if require_all:
            shape_ids = (
                self.feed.shapes.groupby("shape_id").filter(
                    lambda x: all(
                        i in x[TransitNetwork.SHAPES_FOREIGN_KEY].tolist()
                        for i in node_ids
                    )
                )
            ).shape_id.drop_duplicates()
        else:
            shape_ids = self.feed.shapes[
                self.feed.shapes[TransitNetwork.SHAPES_FOREIGN_KEY].isin(node_ids)
            ].shape_id.drop_duplicates()

        # Return pandas.Series of trip_ids
        return self.feed.trips[self.feed.trips.shape_id.isin(shape_ids)].trip_id

    def apply_transit_feature_change(
        self, trip_ids: pd.Series, properties: list, in_place: bool = True
    ) -> Union(None, TransitNetwork):
        """
        Changes the transit attributes for the selected features based on the
        project card information passed

        Args:
            trip_ids : pd.Series
                all trip_ids to apply change to
            properties : list of dictionaries
                transit properties to change
            in_place : bool
                whether to apply changes in place or return a new network

        Returns:
            None
        """
        for i in properties:
            if i["property"] in ["headway_secs"]:
                self._apply_transit_feature_change_frequencies(trip_ids, i, in_place)

            elif i["property"] in ["routing"]:
                self._apply_transit_feature_change_routing(trip_ids, i, in_place)

    def _apply_transit_feature_change_routing(
        self, trip_ids: pd.Series, properties: dict, in_place: bool = True
    ) -> Union(None, TransitNetwork):
        shapes = self.feed.shapes
        stop_times = self.feed.stop_times
        stops = self.feed.stops

        # A negative sign in "set" indicates a traversed node without a stop
        # If any positive numbers, stops have changed
        stops_change = False
        if any(x > 0 for x in properties["set"]):
            # Simplify "set" and "existing" to only stops
            properties["set_stops"] = [str(i) for i in properties["set"] if i > 0]
            if properties.get("existing") is not None:
                properties["existing_stops"] = [
                    str(i) for i in properties["existing"] if i > 0
                ]
            stops_change = True

        # Convert ints to objects
        properties["set_shapes"] = [str(abs(i)) for i in properties["set"]]
        if properties.get("existing") is not None:
            properties["existing_shapes"] = [
                str(abs(i)) for i in properties["existing"]
            ]

        # Replace shapes records
        shape_ids = self.feed.trips[self.feed.trips.trip_id.isin(trip_ids)].shape_id
        for shape_id in shape_ids:
            # Pop the rows that match shape_id
            this_shape = shapes[shapes.shape_id == shape_id]

            # Make sure they are ordered by shape_pt_sequence
            this_shape = this_shape.sort_values(by=["shape_pt_sequence"])

            # Build a pd.DataFrame of new shape records
            new_shape_rows = pd.DataFrame(
                {
                    "shape_id": shape_id,
                    "shape_pt_lat": None,  # FIXME Populate from self.road_net?
                    "shape_pt_lon": None,  # FIXME
                    "shape_osm_node_id": None,  # FIXME
                    "shape_pt_sequence": None,
                    TransitNetwork.SHAPES_FOREIGN_KEY: properties["set_shapes"],
                }
            )

            # If "existing" is specified, replace only that segment
            # Else, replace the whole thing
            if properties.get("existing") is not None:
                # Match list
                nodes = this_shape[TransitNetwork.SHAPES_FOREIGN_KEY].tolist()
                index_replacement_starts = nodes.index(properties["existing_shapes"][0])
                index_replacement_ends = nodes.index(properties["existing_shapes"][-1])
                this_shape = pd.concat(
                    [
                        this_shape.iloc[:index_replacement_starts],
                        new_shape_rows,
                        this_shape.iloc[index_replacement_ends + 1 :],
                    ],
                    sort=False,
                    ignore_index=True,
                )
            else:
                this_shape = new_shape_rows

            # Renumber shape_pt_sequence
            this_shape["shape_pt_sequence"] = np.arange(len(this_shape))

            # Add rows back into shapes
            shapes = pd.concat(
                [shapes[shapes.shape_id != shape_id], this_shape], ignore_index=True
            )

        # Replace stop_times and stops records (if required)
        if stops_change:
            # If node IDs in properties["set_stops"] are not already
            # in stops.txt, create a new stop_id for them in stops
            if any(
                x not in stops[TransitNetwork.STOPS_FOREIGN_KEY].tolist()
                for x in properties["set_stops"]
            ):
                # FIXME
                WranglerLogger.error(
                    "Node ID is used that does not have an existing stop."
                )
                raise ValueError

            # Loop through all the trip_ids
            for trip_id in trip_ids:
                # Pop the rows that match trip_id
                this_stoptime = stop_times[stop_times.trip_id == trip_id]

                # Merge on node IDs using stop_id (one node ID per stop_id)
                this_stoptime = this_stoptime.merge(
                    stops[["stop_id", TransitNetwork.STOPS_FOREIGN_KEY]],
                    how="left",
                    on="stop_id",
                )

                # Make sure the stop_times are ordered by stop_sequence
                this_stoptime = this_stoptime.sort_values(by=["stop_sequence"])

                # Build a pd.DataFrame of new shape records from properties
                new_stoptime_rows = pd.DataFrame(
                    {
                        "trip_id": trip_id,
                        "arrival_time": None,
                        "departure_time": None,
                        "pickup_type": None,
                        "drop_off_type": None,
                        "stop_distance": None,
                        "timepoint": None,
                        "stop_is_skipped": None,
                        TransitNetwork.STOPS_FOREIGN_KEY: properties["set_stops"],
                    }
                )

                # Merge on stop_id using node IDs (many stop_id per node ID)
                new_stoptime_rows = (
                    new_stoptime_rows.merge(
                        stops[["stop_id", TransitNetwork.STOPS_FOREIGN_KEY]],
                        how="left",
                        on=TransitNetwork.STOPS_FOREIGN_KEY,
                    )
                    .groupby([TransitNetwork.STOPS_FOREIGN_KEY])
                    .head(1)
                )  # pick first

                # If "existing" is specified, replace only that segment
                # Else, replace the whole thing
                if properties.get("existing") is not None:
                    # Match list (remember stops are passed in with node IDs)
                    nodes = this_stoptime[TransitNetwork.STOPS_FOREIGN_KEY].tolist()
                    index_replacement_starts = nodes.index(
                        properties["existing_stops"][0]
                    )
                    index_replacement_ends = nodes.index(
                        properties["existing_stops"][-1]
                    )
                    this_stoptime = pd.concat(
                        [
                            this_stoptime.iloc[:index_replacement_starts],
                            new_stoptime_rows,
                            this_stoptime.iloc[index_replacement_ends + 1 :],
                        ],
                        sort=False,
                        ignore_index=True,
                    )
                else:
                    this_stoptime = new_stoptime_rows

                # Remove node ID
                del this_stoptime[TransitNetwork.STOPS_FOREIGN_KEY]

                # Renumber stop_sequence
                this_stoptime["stop_sequence"] = np.arange(len(this_stoptime))

                # Add rows back into stoptime
                stop_times = pd.concat(
                    [stop_times[stop_times.trip_id != trip_id], this_stoptime],
                    ignore_index=True,
                )

        # Replace self if in_place, else return
        if in_place:
            self.feed.shapes = shapes
            self.feed.stops = stops
            self.feed.stop_times = stop_times
        else:
            updated_network = copy.deepcopy(self)
            updated_network.feed.shapes = shapes
            updated_network.feed.stops = stops
            updated_network.feed.stop_times = stop_times
            return updated_network

    def _apply_transit_feature_change_frequencies(
        self, trip_ids: pd.Series, properties: dict, in_place: bool = True
    ) -> Union(None, TransitNetwork):
        freq = self.feed.frequencies

        # Grab only those records matching trip_ids (aka selection)
        freq = freq[freq.trip_id.isin(trip_ids)]

        # Check all `existing` properties if given
        if properties.get("existing") is not None:
            if not all(freq.headway_secs == properties["existing"]):
                WranglerLogger.error(
                    "Existing does not match for at least "
                    "1 trip in:\n {}".format(trip_ids.to_string())
                )
                raise ValueError

        # Calculate build value
        if properties.get("set") is not None:
            build_value = properties["set"]
        else:
            build_value = [i + properties["change"] for i in freq.headway_secs]

        # Update self or return a new object
        q = self.feed.frequencies.trip_id.isin(freq["trip_id"])
        if in_place:
            self.feed.frequencies.loc[q, properties["property"]] = build_value
        else:
            updated_network = copy.deepcopy(self)
            updated_network.loc[q, properties["property"]] = build_value
            return updated_network

    def apply_transit_managed_lane(
        self, trip_ids: pd.Series, node_ids: list, in_place: bool = True
    ) -> Union(None, TransitNetwork):
        # Traversed nodes without a stop should be negative integers
        all_stops = self.feed.stops[TransitNetwork.STOPS_FOREIGN_KEY].tolist()
        node_ids = [int(x) if str(x) in all_stops else int(x) * -1 for x in node_ids]

        self._apply_transit_feature_change_routing(
            trip_ids=trip_ids,
            properties={
                "existing": node_ids,
                "set": RoadwayNetwork.get_managed_lane_node_ids(node_ids),
            },
            in_place=in_place,
        )


class DotDict(dict):
    """
    dot.notation access to dictionary attributes
    Source: https://stackoverflow.com/questions/2352181/how-to-use-a-dot-to-access-members-of-dictionary
    """

    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)
